class RecipeError(Exception):
    def __init__(self, item, message=None, line=None, column=None):
        self.line = line
        self.column = column
        if message is not None:
            if line is not None:
                if column is not None:
                    message += " (at line %i / column %i)" % (line, column)
                else:
                    message += " (at line %i)" % line
            super().__init__(item, message)
        else:
            super().__init__(item)


class EmptyRecipe(RecipeError):
    """Raised if the recipe file is empty"""

    template = "is empty"


class MissingMetaYaml(RecipeError):
    """Raised when FileNotFoundError is encountered

    self.item is NOT a Recipe but a str here
    """

    template = "has missing file `meta.yaml`"


class JinjaRenderFailure(RecipeError):
    """Raised on Jinja rendering problems"""

    template = "failed to render in Jinja2. Error was: %s"


class YAMLRenderFailure(RecipeError):
    """Raised on YAML parsing problems"""

    template = "failed to load YAML. Error was: %s"



import requests
import json
import logging
from conda.models.version import VersionOrder
from conda.exceptions import InvalidVersionSpec
from percy.render.recipe import Package


def get_latest_package_list(
    subdir: str = "linux-64", merge_noarch: bool = True
) -> dict[str, dict]:
    """Get latest packages on defaults

    Args:
        subdir (str, optional): The subdirectory. Defaults to "linux-64".
        merge_noarch (bool, optional): Include noarch versions. Defaults to True.

    Raises:
        Exception: Failed to retrieve repodata

    Returns:
        dict[str,dict]: { pkg_name : { version, build_number, noarch } }
    """

    session = requests.Session()

    # load noarch repodata
    if merge_noarch:
        pkgs_noarch = {}
        repodata_subdir = None
        url = "https://repo.anaconda.com/pkgs/main/noarch/repodata.json"
        response = session.get(url)
        if response.status_code != 200:
            raise Exception("Failed to retrive noarch data.")
        else:
            repodata_subdir = json.loads(response.text)
        for v in repodata_subdir["packages"].values():
            if (
                (v["name"] not in pkgs_noarch)
                or (
                    VersionOrder(pkgs_noarch[v["name"]]["version"])
                    < VersionOrder(v["version"])
                )
                or (
                    VersionOrder(pkgs_noarch[v["name"]]["version"])
                    == VersionOrder(v["version"])
                    and pkgs_noarch[v["name"]]["build_number"] < v["build_number"]
                )
            ):
                pkgs_noarch[v["name"]] = {
                    "version": v["version"],
                    "build_number": v["build_number"],
                    "noarch": True,
                }

    # load subdir repodata
    pkgs_subdir = {}
    repodata_subdir = None
    url = f"https://repo.anaconda.com/pkgs/main/{subdir}/repodata.json"
    response = session.get(url)
    if response.status_code != 200:
        raise Exception("Failed to retrive subdir data.")
    else:
        repodata_subdir = json.loads(response.text)
    for v in repodata_subdir["packages"].values():
        if (
            (v["name"] not in pkgs_subdir)
            or (
                VersionOrder(pkgs_subdir[v["name"]]["version"])
                < VersionOrder(v["version"])
            )
            or (
                VersionOrder(pkgs_subdir[v["name"]]["version"])
                == VersionOrder(v["version"])
                and pkgs_subdir[v["name"]]["build_number"] < v["build_number"]
            )
        ):
            pkgs_subdir[v["name"]] = {
                "version": v["version"],
                "build_number": v["build_number"],
                "noarch": False,
            }

    session.close()

    # merge noarch with subdir
    pkgs_defaults = pkgs_subdir
    if merge_noarch:
        for name, info in pkgs_noarch.items():
            if name in pkgs_defaults and (
                (
                    VersionOrder(info["version"])
                    > VersionOrder(pkgs_defaults[name]["version"])
                )
                or (
                    VersionOrder(info["version"])
                    == VersionOrder(pkgs_defaults[name]["version"])
                    and info["build_number"] > pkgs_defaults[name]["build_number"]
                )
            ):
                logging.info(f"Found newer noarch {name} {pkgs_noarch[name]}")
                pkgs_defaults[name] = pkgs_noarch[name]

    return pkgs_defaults


def compare_package_with_defaults(
    package: Package, defaults_pkgs: dict[str, dict]
) -> dict:
    """Compare a local package with its latest version on defaults

    Args:
        package (Package): Local package
        defaults_pkgs (dict[str,dict]): defaults data

    Returns:
        dict: comparison result
    """
    result = None
    try:
        local_feedstock = package.git_info.name
        local_name = package.name
        local_version = package.version
        local_build_number = int(package.number)
        if local_name in defaults_pkgs:
            defaults_version = defaults_pkgs[local_name]["version"]
            defaults_build_number = int(defaults_pkgs[local_name]["build_number"])
            if (VersionOrder(local_version) < VersionOrder(defaults_version)) or (
                VersionOrder(local_version) == VersionOrder(defaults_version)
                and local_build_number < defaults_build_number
            ):
                result = {
                    "local_feedstock": local_feedstock,
                    "local_version": local_version,
                    "local_build_number": local_build_number,
                    "defaults_version": defaults_version,
                    "defaults_build_number": defaults_build_number,
                }
    except InvalidVersionSpec as exc:
        logging.error(f"Exception: {local_name} {exc}")
    except ValueError as exc:
        logging.error(f"Exception: {local_name} {exc}")
    return result



""" Find what to rebuild when updating a feedstock.
    Create corresponding tickets in Jira.
"""

import percy.render.aggregate as aggregate
import argparse
from pathlib import Path
import requests
import json
import itertools
import logging
import os
from jira.client import JIRA
from jira.exceptions import JIRAError

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)

perseverance_jira = None
perseverance_fields_map = None


def add_issue(login, token, dry_run, current_epic_key, target_pkg, pkg):
    """
    Add a new issue to perseverance jira.
    """
    global perseverance_jira
    global perseverance_fields_map

    if not perseverance_jira:
        perseverance_jira = JIRA(
            "https://anaconda.atlassian.net/", basic_auth=(login, token)
        )
        allfields = perseverance_jira.fields()
        perseverance_fields_map = {field["name"]: field["id"] for field in allfields}

    logging.info(f"NEW {pkg}")

    issue_fields = {
        "project": {"key": "PKG"},
        "summary": f"{pkg} rebuilt against {target_pkg}",
        "description": f"{pkg} rebuilt against {target_pkg}",
        "issuetype": {"name": "Task"},
        perseverance_fields_map["Package Name"]: pkg,
    }
    if current_epic_key:
        issue_fields[perseverance_fields_map["Epic Link"]] = current_epic_key
    try:
        # create issue
        if not dry_run:
            perseverance_jira.create_issue(fields=issue_fields, prefetch=True)
        else:
            logging.info(issue_fields)

    except JIRAError as err:
        logging.error(f"Failed to add issue {pkg}. {err.text}")


def depends_on_target(repo_package, target):
    for v in repo_package["depends"]:
        if v.startswith(f"{target} "):
            return v
    return None


def process_repodata(subdir, targets):
    session = requests.Session()
    pkgs = []
    is_noarch = subdir == "noarch"

    # load subdir repodata
    repodata_subdir = None
    url = f"https://repo.anaconda.com/pkgs/main/{subdir}/repodata.json"
    response = session.get(url)
    if response.status_code != 200:
        raise Exception()
    else:
        repodata_subdir = json.loads(response.text)
    for v in repodata_subdir["packages"].values():
        for target in targets:
            spec_str = depends_on_target(v, target)
            if spec_str:
                pkgs.append(
                    {
                        "name": v["name"],
                        "spec": spec_str,
                        "version": v["version"],
                        "build_number": v["build_number"],
                        "noarch": is_noarch,
                    }
                )

    pkgs = sorted(pkgs, key=lambda v: v["name"])

    session.close()

    return pkgs


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="aggregate_issue_finder",
        description="Find issues in aggregate pinned feedstocks.",
    )
    parser.add_argument(
        "-a",
        "--aggregate",
        type=Path,
        help="aggregate path",
        default="/Users/cbousseau/work/recipes/aggregate2/",
    )
    parser.add_argument("--epic", type=str, help="Epic key")
    parser.add_argument(
        "--target", type=str, help="Target feedstock to rebuild against"
    )
    parser.add_argument(
        "--dry_run", help="If set, no actual sync is done.", action="store_true"
    )
    parser.add_argument(
        "--login",
        type=str,
        help="Jira login (if not given, env var JIRA_LOGIN is used)",
    )
    parser.add_argument(
        "--token",
        type=str,
        help="Jira token (if not given, env var JIRA_TOKEN is used)",
    )

    return parser


class JiraSync:
    def __init__(self, login, token, dry_run=False):
        """
        If dry_run is True, no calls are made to update Jira.
        """
        self.dry_run = dry_run
        self.perseverance_project = "PKG"
        self.perseverance_jira = JIRA(
            "https://anaconda.atlassian.net/", basic_auth=(login, token)
        )
        allfields = self.perseverance_jira.fields()
        self.perseverance_fields_map = {
            field["name"]: field["id"] for field in allfields
        }

    def sync(self, epic_key, target, stages):
        """
        Add changes from input json to perseverance jira
        """

        # read epic tasks from jira
        epic_tasks = []
        total = 0
        while True:
            issues = self.perseverance_jira.search_issues(
                f'project = {self.perseverance_project} and "Epic Link" = {epic_key}',
                fields="*all",
                startAt=total,
            )
            got = len(issues)
            if got == 0:
                break
            total += got
            epic_tasks.extend(
                [
                    getattr(
                        issue.fields, self.perseverance_fields_map["Package Name"], ""
                    ).lower()
                    for issue in issues
                ]
            )

        # add packages to jira
        n_stages = len(stages)
        for i, stage in enumerate(stages):
            for feedstock in stage:
                package_name = next(iter(feedstock.packages)).lower()
                if package_name not in epic_tasks:
                    logging.info(
                        f"No existing ticket {package_name} : creating new ticket"
                    )
                    issue_key = f"{i+1:03}/{n_stages:03} {feedstock.name}"
                    desc = f"{issue_key} rebuild against {target}"
                    issue_fields = {
                        "project": {"key": "PKG"},
                        "summary": desc,
                        "description": desc,
                        "issuetype": {"name": "Task"},
                        self.perseverance_fields_map["Package Name"]: package_name,
                    }
                    issue_fields[self.perseverance_fields_map["Epic Link"]] = epic_key
                    try:
                        # create issue
                        if not self.dry_run:
                            self.perseverance_jira.create_issue(
                                fields=issue_fields, prefetch=True
                            )
                        else:
                            logging.info(issue_fields)

                    except JIRAError as err:
                        logging.error(
                            f"Failed to add issue {feedstock.name}. {err.text}"
                        )
                else:
                    logging.info(f"Existing ticket {package_name} : skip")


def print_build_order(buildout):
    stages = [
        list(result)
        for key, result in itertools.groupby(buildout, key=lambda f: f.weight)
    ]
    for i, stage in enumerate(stages):
        for feedstock in stage:
            print(f"{i:03} {feedstock.name:30} {list(feedstock.packages.keys())}")


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()

    subdirs = [
        "osx-arm64",
        "osx-64",
        "win-64",
        "linux-64",
        "linux-aarch64",
        "linux-ppc64le",
        "linux-s390x",
    ]

    python = [
        "3.8",
        "3.9",
        "3.10",
        "3.11",
    ]

    # load aggregate
    aggregate_repo = aggregate.Aggregate(args.aggregate)
    for subdir, pyver in itertools.product(subdirs, python):
        others = {"r_implementation": "r-base"}
        if subdir.startswith("win-"):
            others["rust_compiler"] = "rust"
        aggregate_repo.load_local_feedstocks(subdir, pyver, others)

    # load repodata
    pkg_targets = aggregate_repo.feedstocks[args.target].packages.keys()
    main_pkgs = set()
    for subdir in subdirs + ["noarch"]:
        results = process_repodata(subdir, pkg_targets)
        main_pkgs.update(set([r["name"] for r in results]))
        with open(f"./{args.target}_{subdir}.json", "w") as f:
            json.dump(results, f, indent=4)

    # get feedstock build order
    buildout = aggregate_repo.get_depends_build_order(
        [], [args.target], [], [], [], False
    )

    # print build order
    print("\n\nDownstream build order:")
    stages = [
        list(result)
        for key, result in itertools.groupby(buildout, key=lambda f: f.weight)
    ]
    all_packages = set()
    for i, stage in enumerate(stages):
        for feedstock in stage:
            all_packages.update(set(feedstock.packages.keys()))
            print(f"{i:03} {feedstock.name:30} {list(feedstock.packages.keys())}")
    for pkg in main_pkgs.difference(all_packages):
        print(f"Missing from build order: {pkg}")

    # create jira issues
    login = args.login
    if not args.login:
        login = os.environ.get("JIRA_LOGIN")
    token = args.token
    if not args.token:
        token = os.environ.get("JIRA_TOKEN")
    syncer = JiraSync(login, token, args.dry_run)
    syncer.sync(args.epic, args.target, stages)



"""Test installation of all packages listed in python_full_package_list.yaml

Usage:
    pytest -n auto \
      --junit-xml="test_install.xml" \
      --html="test_install.html" \
      --self-contained-html \
      test_install.py
"""

import json
import pytest
import subprocess
import yaml


def dry_run(
    solver="libmamba",
    channels=["defaults"],
    override_channels=False,
    packages=[],
):
    """Dry-run an environment and return it's json result."""
    cmd = f"conda create -n test_env --dry-run --json --experimental-solver={solver}"
    for channel in channels:
        cmd += f" -c {channel}"
    if override_channels:
        cmd += " --override-channels"
    cmd = cmd.split(" ") + [f"'{p}'" for p in packages]
    json_result = {}
    retry = 5
    while retry > 0:
        retry = retry - 1
        result = subprocess.run(cmd, capture_output=True, text=True)
        json_result = {}
        try:
            json_result.update(json.loads(result.stdout))
            json_result["__command__"] = {"cmd": " ".join(cmd)}
            json_result["__result__"] = result.returncode
            break
        except json.decoder.JSONDecodeError:
            json_result["success"] = False
            json_result["error"] = "json.decoder.JSONDecodeError"
            json_result["stdout"] = result.stdout
            json_result["stderr"] = result.stderr
            json_result["__command__"] = {"cmd": " ".join(cmd)}
            json_result["__result__"] = result.returncode
    return json_result


def dry_run_assert(solver, packages, channels):
    """Dry-run an environment and assert that it succeeds."""
    json_result = dry_run(
        solver=solver,
        channels=channels,
        packages=packages,
    )
    success = json_result.get("success", False)
    if not success:
        message = json_result["error"]
        message += "\nTo reproduce locally:"
        message += f'\n{json_result["__command__"]["cmd"]}'
        message += f"\n\n{json.dumps(json_result, indent=4)}"
    assert json_result.get("success", False), message


def data_source(filename):
    import os

    print(f"data_source({filename=}", os.getcwd())  # XXX
    with open(filename) as f:
        return yaml.safe_load(f).items()


@pytest.mark.parametrize(
    "package, channels",
    data_source("./python_full_package_list.yaml"),
)
def test_install(package, channels):
    packages = [f"{package}", "python[version=3.11]"]
    dry_run_assert("classic", packages, channels)



""" Generate build scripts to build all packages depending on python.
"""

import percy.render.aggregate as aggregate
from config import block_list
from config import extras_versions as extras
import argparse
from pathlib import Path
from itertools import groupby
import requests
import json
import os
import shutil
import yaml


def get_repodata_package_list(subdir, python_ref, include_noarch=False):
    def match_lower(mrv, rv, v):
        irv = int(rv) - 1
        hlp = ">={}.{}".format(mrv, irv)
        if v.find(hlp) >= 0:
            return True
        irv = irv - 1
        while irv > -1:
            hlp = ">={}.{}".format(mrv, irv)
            if v.find(hlp) >= 0:
                hlp += ","
                return v.find(hlp) < 0
            hlp = ">{}.{}".format(mrv, irv)
            if v.find(hlp) >= 0:
                hlp += ","
                return v.find(hlp) < 0
            irv = irv - 1
        return False

    def depends_on_python(repo_package, python_ref):
        buildstr = repo_package["build"]
        pyver = f"py{python_ref.replace('.','')}"
        if "py_" in buildstr or "pyh" in buildstr or pyver in buildstr:
            return True
        repo_package["depends"]

        hlp = ">=" + python_ref + ",<" + python_ref[0] + "."
        hlp2 = "=" + python_ref
        next_minor = str(int(python_ref.split(".")[-1]) + 1)
        for v in repo_package["depends"]:
            if v.find("python ") >= 0:
                if v.find(hlp) >= 0 or match_lower(python_ref[0], next_minor, v):
                    return True
            elif v.find(hlp2) >= 0:
                return True
            elif v == "python":
                return True

        return False

    session = requests.Session()

    repodata_subdir = None
    url = f"https://repo.anaconda.com/pkgs/main/{subdir}/repodata.json"
    response = session.get(url)
    if response.status_code != 200:
        raise Exception()
    else:
        repodata_subdir = json.loads(response.text)

    pkgs = set(
        v["name"]
        for v in repodata_subdir["packages"].values()
        if (depends_on_python(v, python_ref))
    )

    if include_noarch:
        repodata_noarch = None
        url = "https://repo.anaconda.com/pkgs/main/noarch/repodata.json"
        response = session.get(url)
        if response.status_code != 200:
            raise Exception()
        else:
            repodata_noarch = json.loads(response.text)
        pkgs.update(
            set(
                v["name"]
                for v in repodata_noarch["packages"].values()
                if (depends_on_python(v, python_ref))
            )
        )

    return sorted(list(pkgs))


def gen_python_build_order(
    aggregate_path, subdir, python_ref, python_target, numpy_target, croot, channel
):
    # Helper to generate script names.
    def _script(script_name_ending):
        return f"./{subdir}/python_{python_target}_{subdir}_{script_name_ending}"

    if croot is None:
        if not subdir.startswith("win-"):
            croot = f"../ci_py{python_target.replace('.','')}/"
        else:
            croot = f"c:/ci_{python_target.replace('.','')}/"

    os.makedirs(f"./{subdir}/", exist_ok=True)

    # load repodata package list
    repodata_package_list = sorted(get_repodata_package_list(subdir, python_ref, False))
    repodata_package_list_with_noarch = sorted(
        get_repodata_package_list(subdir, python_ref, True)
    )

    # load aggregate
    aggregate_repo = aggregate.Aggregate(aggregate_path)
    others = {"r_implementation": "r-base"}
    if subdir.startswith("win-"):
        others["rust_compiler"] = "rust"
        aggregate_repo.load_local_feedstocks(subdir, python_ref, others)
    else:
        aggregate_repo.load_local_feedstocks(subdir, python_ref, others)

    # write package to feedstock mapping file - for reference
    with open(f"./{subdir}/map_to_p.yaml", "w") as f:
        yaml.dump(aggregate_repo.package_to_feedstock_path(), f)

    # get feedstock build order
    allow_list = repodata_package_list  # only packages already in subdir defaults
    # only packages already in subdir and noarch defaults
    allow_list_noarch = repodata_package_list_with_noarch
    python_buildout = aggregate_repo.get_depends_build_order(
        [], [], ["python"], allow_list, block_list, True
    )

    # compare local aggregate package list to repodata package list - for reference
    aggregate_package_list = []
    for feedstock in python_buildout:
        for pkg_name in feedstock.packages.keys():
            aggregate_package_list.append(pkg_name)
    aggregate_package_list = sorted(aggregate_package_list)
    with open(
        f"./{subdir}/python_{python_target}_{subdir}_package_list.yaml", "w"
    ) as f:
        yaml.dump(aggregate_package_list, f)
    repodata_package_list = sorted(get_repodata_package_list(subdir, python_ref))
    with open(f"./{subdir}/python_{python_ref}_{subdir}_package_list.yaml", "w") as f:
        yaml.dump(repodata_package_list, f)
    with open(
        f"./{subdir}/python_{python_ref}_{subdir}_package_list_missing.yaml", "w"
    ) as f:
        yaml.dump(
            sorted(list(set(repodata_package_list) - set(aggregate_package_list))), f
        )
    with open(
        f"./{subdir}/python_{python_ref}_{subdir}_package_list_new.yaml", "w"
    ) as f:
        yaml.dump(
            sorted(list(set(aggregate_package_list) - set(repodata_package_list))), f
        )

    # also export subdir + noarch list
    python_buildout_with_noarch = aggregate_repo.get_depends_build_order(
        [], [], ["python"], allow_list_noarch, block_list, False
    )
    aggregate_package_list_with_noarch = []
    for feedstock in python_buildout_with_noarch:
        for pkg_name in feedstock.packages.keys():
            aggregate_package_list_with_noarch.append(pkg_name)
    aggregate_package_list_with_noarch = sorted(aggregate_package_list_with_noarch)
    with open(f"./{subdir}/python_full_package_list.yaml", "w") as f:
        yaml.dump(
            dict(
                map(lambda e: (e, [croot, channel]), aggregate_package_list_with_noarch)
            ),
            f,
        )
    shutil.copy(Path(__file__).parent / "test_install.py", f"./{subdir}/")

    # write stage build scripts
    stages = [
        list(result) for key, result in groupby(python_buildout, key=lambda f: f.weight)
    ]
    n_stages = len(stages)

    if not subdir.startswith("win-"):
        build = " ".join(
            [
                "conda-build",
                "--python={python_ver}",
                "--numpy={numpy_ver}",
                "--croot={croot}",
                "-c {channel}",
                "--use-local",
                "--no-test",
                "{feedstockpath}",
                ">d 2>&1",
                "&& rm -f d",
                "&& ( echo done >>{feedstockname}.mark )",
            ]
        )
        show_status = " && ".join(
            [
                '( echo "{feedstockname}" >>failed.{stage} )',
                '( echo "{feedstockname}" >>errors.dump )',
                "( cat d >>errors.dump )",
                "cat d",
                "rm -f d",
            ]
        )
        build_feedstock_template = f"""
if [[ ! -f {{feedstockname}}.mark ]]; then
    if [[ -d {{feedstockname}} ]]; then
        ( ( {build} ) || ( {show_status} ) )
    else
        echo '{{feedstockname}} not present' >>failed.{{stage}}
    fi
fi

"""

        with open(_script("build_all.sh"), "w") as g:
            g.write("#!/bin/bash\n")
            g.write("set -x\n")
            for i, stage in enumerate(stages, 1):
                fname = _script(f"stage_{i:02}_of_{n_stages:02}.sh")
                g.write(f"{fname.replace(subdir+'/', '')}\n")
                with open(fname, "w") as f:
                    f.write("#!/bin/bash\n")
                    f.write("CONDA_ADD_PIP_AS_PYTHON_DEPENDENCY=0\n")
                    for feedstock in stage:
                        f.write(
                            build_feedstock_template.format(
                                feedstockname=feedstock.name,
                                feedstockpath=feedstock.path,
                                python_ver=python_target,
                                numpy_ver=numpy_target,
                                croot=croot,
                                channel=channel,
                                stage=i,
                            )
                        )

                        if feedstock.name in extras:
                            f.write(
                                build_feedstock_template.format(
                                    feedstockname=extras[feedstock.name],
                                    feedstockpath=f"./{extras[feedstock.name]}",
                                    python_ver=python_target,
                                    numpy_ver=numpy_target,
                                    croot=croot,
                                    channel=channel,
                                    stage=i,
                                )
                            )
                    g.write(f'echo "Done stage {i:02} of {n_stages:02}"\n')
            g.write('echo "Done"\n')

        # copy test script
        shutil.copy(Path(__file__).parent / "test_packages.sh", f"./{subdir}/")

        # copy tally script
        shutil.copy(Path(__file__).parent / "tally.sh", f"./{subdir}/")

    else:
        # Build the command template string.
        build = " ".join(
            [
                "conda-build",
                "--python={python_ver}",
                "--numpy={numpy_ver}",
                "--croot={croot}",
                "-c {channel}",
                "--use-local",
                "--no-test",
                "{feedstockpath}",
            ]
        )
        show_status = 'echo "{feedstockname}" >>failed.{stage}'
        quit = 'cmd /K "exit /b 0"'
        build_feedstock_template = f"\n{build} || {show_status} || {quit}\n"

        with open(_script("build_all.bat"), "w") as g:
            for i, stage in enumerate(stages, 1):
                script = _script(f"stage_{i:02}_of_{n_stages:02}.bat")
                fname = script.rsplit("/", 1)[1]
                g.write(f"call {fname}\n")
                with open(script, "w") as f:
                    f.write("set CONDA_ADD_PIP_AS_PYTHON_DEPENDENCY=0\r\n")
                    for feedstock in stage:
                        f.write(
                            build_feedstock_template.format(
                                feedstockname=feedstock.name,
                                feedstockpath=feedstock.path,
                                python_ver=python_target,
                                numpy_ver=numpy_target,
                                croot=croot,
                                channel=channel,
                                stage=i,
                            )
                        )
                        if feedstock.name in extras:
                            f.write(
                                build_feedstock_template.format(
                                    feedstockname=extras[feedstock.name],
                                    feedstockpath=f"./{extras[feedstock.name]}",
                                    python_ver=python_target,
                                    numpy_ver=numpy_target,
                                    croot=croot,
                                    channel=channel,
                                    stage=i,
                                )
                            )
                    g.write(f'echo "Done stage {i:02} of {n_stages:02}"\n')
            g.write('echo "Done"\n')

        # copy test script
        shutil.copy("test_packages.bat", f"./{subdir}/")


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="gen_python_build_order",
        description="Build order for a new python release.",
    )
    parser.add_argument(
        "-a",
        "--aggregate",
        type=Path,
        help="aggregate path",
        default="/Users/cbousseau/work/recipes/aggregate2/",
    )
    parser.add_argument(
        "-s", "--subdir", type=str, help="Architecture", default="linux-64"
    )
    parser.add_argument(
        "-r", "--python-ref", type=str, help="Reference python version", default="3.11"
    )
    parser.add_argument(
        "-p", "--python-target", type=str, help="Target python version", default="3.12"
    )
    parser.add_argument(
        "-n", "--numpy-target", type=str, help="Target numpy version", default="1.25"
    )
    parser.add_argument(
        "-c",
        "--channel",
        type=str,
        help="Python release channel",
        default="ad-testing/label/py312",
    )
    parser.add_argument("-q", "--croot", type=str, help="croot")

    return parser


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    gen_python_build_order(
        args.aggregate,
        args.subdir,
        args.python_ref,
        args.python_target,
        args.numpy_target,
        args.croot,
        args.channel,
    )



from pathlib import Path
import logging
import pytest
from percy.examples.preinstall.dry_run import dry_run
import percy.render.recipe as recipe

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)

def render(feedstock_path, aggregate_path, arch, python):
    logging.debug(f'Render feedstock {feedstock_path} for { arch }')

    recipe_path = feedstock_path / 'recipe' / 'meta.yaml'
    others = {"r_implementation": "r-base"}
    if arch.startswith("win-"):
        others["rust_compiler"] = "rust"
    render_results = recipe.render(recipe_path, arch, python, others)

    return render_results

def check_deps(
        feedstock_path,
        aggregate_path,
        arch='linux-aarch64',
        python='3.10',
        channels=None,
        extras=None,
    ):
    # Set defaults to function-scoped locals if not specified as parameters.
    if not channels:
        channels = ['default']
    if not extras:
        extras = []

    logging.debug(f'Render feedstock {feedstock_path} for { arch }')

    recipe_path = feedstock_path / 'recipe' / 'meta.yaml'
    others = {"r_implementation": "r-base"}
    if arch.startswith("win-"):
        others["rust_compiler"] = "rust"
    render_results = recipe.render(recipe_path, arch, python, others)
    unresolved_deps = {}
    for render_result in render_results:
        meta = render_result.meta

        if meta:
            try:
                logging.debug('Check run environment')
                name = meta.get('package', {}).get('name', 'unknown')
                pkg_reqs = []
                if python != 'nopy':
                    pkg_reqs.extend([f'python={python}'])
                pkg_reqs.extend(extras)
                requirements = meta.get('requirements', {})
                if requirements is not None:
                    for e in ['build', 'host', 'run', 'run_constrained']:
                        req = requirements.get(e, [])
                        if req is not None:
                            pkg_reqs.extend(req)
                test = meta.get('test', {})
                if test is not None:
                    req = test.get('requires', [])
                    if req is not None:
                        pkg_reqs.extend(req)
                outputs = meta.get('outputs', [])
                if outputs:
                    for output in outputs:
                        name = output.get('name', '')
                        output_pkg_reqs = pkg_reqs
                        requirements = output.get('requirements', {})
                        if requirements is not None:
                            for e in ['build', 'host', 'run', 'run_constrained']:
                                req = requirements.get(e, [])
                                if req is not None:
                                    pkg_reqs.extend(req)
                        test = output.get('test', {})
                        if test is not None:
                            req = test.get('requires', [])
                            if req is not None:
                                pkg_reqs.extend(req)
                        unresolved_deps[name] = dry_run(
                            subdir=arch,
                            channels=channels,
                            packages=output_pkg_reqs,
                        )
                else:
                    unresolved_deps[name] = dry_run(
                        subdir=arch,
                        channels=channels,
                        packages=pkg_reqs,
                    )
        
            except:
                print(meta)
                raise

    return unresolved_deps


@pytest.fixture(scope='session')
def feedstock_path(pytestconfig):
    return Path(pytestconfig.getoption('feedstock')).expanduser().absolute()


@pytest.fixture(scope='session')
def aggregate_path(pytestconfig):
    return Path(pytestconfig.getoption('aggregate')).expanduser().absolute()


@pytest.fixture(scope='session')
def channels(pytestconfig):
    return pytestconfig.getoption('channels')


@pytest.fixture(scope='session')
def extras(pytestconfig):
    return pytestconfig.getoption('extras')


@pytest.hookimpl
def pytest_generate_tests(metafunc):
    if 'arch' in metafunc.fixturenames:
        values = metafunc.config.option.arch
        metafunc.parametrize('arch', values)
    recipe_path = Path(
        metafunc.config.option.feedstock
    ).expanduser().absolute() / 'recipe' / 'meta.yaml'
    with open(recipe_path, 'r') as stream:
        contents = stream.read()
        if 'python' in contents:
            values = metafunc.config.option.python
            metafunc.parametrize('python', values)
        else:
            metafunc.parametrize('python', ['nopy'])


def test_deps(feedstock_path, aggregate_path, arch, python, channels, extras):
    if not (arch == 'osx-arm64' and python == '3.7'):
        results = check_deps(feedstock_path, aggregate_path,
                            arch, python, channels, extras)
        for name, problems in results.items():
            success = problems.get("success", False)
            if not success:
                print(problems)
            assert success



import os
import subprocess
import sys
import json
import argparse

def dry_run(
    subdir="linux-64",
    glibc=None,
    solver="libmamba",
    channels=["defaults"],
    override_channels=False,
    packages=[],
):
    """ Dry-run an environment and return it's json result.
    """
    cmd = f"conda create -n test_env --dry-run --json --experimental-solver={solver}"
    for channel in channels:
        cmd += f" -c {channel}"
    if subdir.startswith("win-"):
        cmd += " -c msys2"
    if override_channels:
        cmd += " --override-channels"
    cmd = cmd.split(" ") + [f"'{p}'" for p in packages]
    my_env = os.environ.copy()
    conda_env = {}
    conda_env["CONDA_SUBDIR"] = f"{subdir}"
    if subdir.startswith("linux-"):
        if glibc is None:
            if subdir == "linux-aarch64":
                glibc = "2.19"
            else:
                glibc = "2.17"
        conda_env["CONDA_OVERRIDE_GLIBC"] = glibc
    my_env.update(conda_env)
    result = subprocess.run(cmd, capture_output=True, text=True, env=my_env)
    json_result = json.loads(result.stdout)
    json_result["__command__"] = {"cmd": " ".join(cmd), "env": conda_env}
    return json_result

def dry_run_assert(subdir, packages):
    """ Dry-run an environment and assert that it succeeds.
    """
    json_result = dry_run(
        subdir=subdir,
        channels=["defaults"],
        override_channels=True,
        packages=packages,
    )
    success = json_result.get("success", False)
    if not success:
        message = json_result["error"]
        message += "\nTo reproduce locally:"
        for k, v in json_result["__command__"]["env"].items():
            message += f"\nexport {k}={v}"
        message += f'\n{json_result["__command__"]["cmd"]}'
        message += f"\n\n{json.dumps(json_result, indent=4)}"
    assert json_result.get("success", False), message

def create_parser() -> argparse.ArgumentParser:
    """ Create command line parser.
    """
    parser = argparse.ArgumentParser(
        prog="test_env",
        description="Dry-run an environment.",
    )
    parser.add_argument(
        "-a", "--subdir", type=str, help="Subdir: e.g. linux-64", default="linux-64"
    )
    parser.add_argument("-g", "--glibc", type=str, help="glibc version", default=None)
    parser.add_argument(
        "-s",
        "--solver",
        type=str,
        help="Solver",
        choices=["classic", "libmamba"],
        default="libmamba",
    )
    parser.add_argument(
        "-c", "--channel", type=str, action="append", default=["defaults"]
    )
    parser.add_argument("--override-channels", action="store_true")
    parser.add_argument(
        "packages",
        type=str,
        help="Matchspec list. e.g: python[version=3.9] anaconda=2022.10",
        nargs="+",
    )
    parser.set_defaults(override_channels=False)

    return parser


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    json_result = dry_run(
        args.subdir,
        args.glibc,
        args.solver,
        args.channel,
        args.override_channels,
        args.packages,
    )
    print(json.dumps(json_result, indent=4))
    if json_result.get("success", False):
        sys.exit(0)
    else:
        sys.exit(1)



print("import: 'percy.render.recipe'")
import percy.render.recipe

print("import: 'percy.render.aggregate'")
import percy.render.aggregate

print("import: 'percy.repodata.repodata'")
import percy.repodata.repodata




from pathlib import Path
import click
import os
import sys
import functools
import yaml
from itertools import groupby
import percy.render.aggregate
import percy.repodata.repodata


def get_configured_aggregate(cmd_line=None):
    # command line has highest precedence
    if cmd_line:
        return cmd_line
    # environment variable
    path = os.getenv("ABS_AGGREGATE")
    if path:
        return path
    # look through ancestor directories
    cwd = Path(os.getcwd())
    path = cwd
    while path != path.parent:
        if (path / "manifest.yaml").exists() or (
            path / "make-mixed-crlf-patch.py"
        ).exists():
            return path
        path = path.parent
    # look in well known locations
    if (cwd / "aggregate" / ".git").exists():
        return cwd / "aggregate"
    return cwd


def load_aggregate(obj, subdir, python, others):
    aggregate_path = obj["aggregate_directory"]
    aggregate_repo = percy.render.aggregate.Aggregate(aggregate_path)
    if not others:
        others = {}
        if subdir.startswith("win-"):
            others["rust_compiler"] = "rust"
    others["r_implementation"] = "r-base"
    aggregate_repo.load_local_feedstocks(subdir, python, others)
    return aggregate_repo


def print_build_order(buildout):
    stages = [list(result) for key, result in groupby(buildout, key=lambda f: f.weight)]
    for i, stage in enumerate(stages):
        for feedstock in stage:
            print(f"{i:03} {feedstock.name:30} {list(feedstock.packages.keys())}")


def base_options(f):
    @click.option(
        "--subdir",
        "-s",
        type=str,
        multiple=False,
        default="linux-64",
        help="Architecture. E.g. -s linux-64",
    )
    @click.option(
        "--python",
        "-p",
        type=str,
        multiple=False,
        default="3.10",
        help="Python version. E.g. -p 3.10",
    )
    @click.option(
        "--others",
        "-k",
        type=(str, str),
        multiple=True,
        default={},
        help="Additional key values (e.g. -k blas_impl openblas)",
    )
    @functools.wraps(f)
    def wrapper_base_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_base_options


def order_options(f):
    @click.option(
        "--groups",
        "-g",
        type=str,
        multiple=True,
        default=[],
        help="List of groups",
    )
    @click.option(
        "--feedstocks",
        "-f",
        type=str,
        multiple=True,
        default=[],
        help="List of feedstocks",
    )
    @click.option(
        "--packages",
        "-pkg",
        type=str,
        multiple=True,
        default=[],
        help="List of packages",
    )
    @click.option(
        "--drop_noarch",
        type=bool,
        is_flag=True,
        show_default=True,
        default=False,
        multiple=False,
        help="Drop noarch",
    )
    @functools.wraps(f)
    def wrapper_order_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_order_options


@click.group(short_help="Commands for operating on aggregates.")
@click.option(
    "--aggregate", "-a", metavar="DIRECTORY", help="Aggregate directory to operate on."
)
@click.pass_context
def aggregate(ctx, aggregate):
    """Commands that operate on aggregates.

    An aggregate is a collection of recipes.

    The aggregate directory is configured from (in order of precedence):

    \b
    1. --aggregate command line option if present
    2. ABS_AGGREGATE env variable if set
    3. The current directory (or an ancestor of it)
    """
    ctx.ensure_object(dict)
    print(Path(get_configured_aggregate(aggregate)))
    ctx.obj["aggregate_directory"] = Path(get_configured_aggregate(aggregate))


@aggregate.command(short_help="Print downstream build order")
@click.pass_obj
@base_options
@order_options
@click.option(
    "--allow_list",
    type=str,
    multiple=True,
    default=[],
    help="List of packages",
)
@click.option(
    "--block_list",
    type=str,
    multiple=True,
    default=[],
    help="List of feedstocks",
)
def downstream(
    obj,
    subdir,
    python,
    others,
    groups,
    feedstocks,
    packages,
    allow_list,
    block_list,
    drop_noarch,
):
    """Prints build order of feedstock downstream dependencies"""

    # load aggregate
    aggregate_repo = load_aggregate(obj, subdir, python, others)

    # get feedstock build order
    buildout = aggregate_repo.get_depends_build_order(
        groups, feedstocks, packages, allow_list, block_list, drop_noarch
    )

    # print build order
    order = " ".join([
        f"groups:{groups}",
        f"feedstocks:{feedstocks}",
        f"packages:{packages}",
        f"allow_list:{allow_list}",
        f"block_list:{block_list}",
    ])
    print(f"\n\nDownstream build order ({order}):")
    print_build_order(buildout)


@aggregate.command(short_help="Print upstream build order")
@click.pass_obj
@base_options
@order_options
def upstream(obj, subdir, python, others, groups, feedstocks, packages, drop_noarch):
    """Prints build order of feedstock upstream dependencies"""

    # load aggregate
    aggregate_repo = load_aggregate(obj, subdir, python, others)

    # get feedstock build order
    buildout = aggregate_repo.get_build_order(
        groups, feedstocks, packages, drop_noarch, False
    )

    # print build order
    order = " ".join([
        f"groups:{groups}",
        f"feedstocks:{feedstocks}",
        f"packages:{packages}",
        f"drop_noarch:{drop_noarch}",
    ])
    print(f"\n\nUpstream build order ({order}):")
    print_build_order(buildout)


@aggregate.command(short_help="Print build order")
@click.pass_obj
@base_options
@order_options
def order(obj, subdir, python, others, groups, feedstocks, packages, drop_noarch):
    """Prints build order of specified feedstocks"""

    # load aggregate
    aggregate_repo = load_aggregate(obj, subdir, python, others)

    # get feedstock build order
    if not groups and not feedstocks and not packages:
        feedstocks = aggregate_repo.feedstocks.keys()
    buildout = aggregate_repo.get_build_order(
        groups, feedstocks, packages, drop_noarch, True
    )

    # print build order
    order = " ".join([
        f"groups:{groups}",
        f"feedstocks:{feedstocks}",
        f"packages:{packages}",
        f"drop_noarch:{drop_noarch}",
    ])
    print(f"\n\nBuild order ({order})):")
    print_build_order(buildout)


@aggregate.command(short_help="Print outdated with defaults")
@click.pass_obj
@base_options
@click.option(
    "--missing_local",
    type=bool,
    is_flag=True,
    show_default=True,
    default=False,
    multiple=False,
    help="Identify packages from defaults not pinned on aggregate.",
)
@click.option(
    "--missing_defaults",
    type=bool,
    is_flag=True,
    show_default=True,
    default=False,
    multiple=False,
    help="Identify packages from aggregate not on defaults",
)
def outdated(obj, subdir, python, others, missing_local, missing_defaults):
    """Prints outdated with defaults"""

    results = {}

    # load aggregate
    aggregate_repo = load_aggregate(obj, subdir, python, others)

    # load defaults
    defaults_pkgs = percy.repodata.repodata.get_latest_package_list(subdir, True)

    # compare aggregate with defaults
    for local_name, package in aggregate_repo.packages.items():
        result = percy.repodata.repodata.compare_package_with_defaults(
            package, defaults_pkgs
        )
        if result:
            results[local_name] = result

    # find missing from local
    if missing_local:
        for name in defaults_pkgs.keys() - aggregate_repo.packages.keys():
            results[name] = {
                "local_feedstock": None,
                "local_version": None,
                "local_build_number": None,
                "defaults_version": defaults_pkgs[name]["version"],
                "defaults_build_number": int(defaults_pkgs[name]["build_number"]),
            }

    # find missing from defaults
    if missing_defaults:
        for name in aggregate_repo.packages.keys() - defaults_pkgs.keys():
            results[name] = {
                "local_feedstock": package.git_info.name,
                "local_version": package.version,
                "local_build_number": int(package.number),
                "defaults_version": None,
                "defaults_build_number": None,
            }

    # dump result
    def noop(self, *args, **kw):
        pass

    yaml.emitter.Emitter.process_tag = noop
    yaml.dump(
        results,
        sys.stdout,
        default_flow_style=False,
        indent=2,
    )



""" Generate build scripts to build all packages depending on python.
"""

import percy.render.aggregate as aggregate
from itertools import groupby
import os
import yaml

def gen_blts_build_order(
    aggregate_path, subdir, python_ref, packages
):

    os.makedirs(f"./{subdir}/", exist_ok=True)

    # load aggregate
    aggregate_repo = aggregate.Aggregate(aggregate_path)
    others = {"r_implementation": "r-base"}
    others["blas_impl"] = "openblas"
    if subdir.startswith("win-"):
        others["rust_compiler"] = "rust"
        aggregate_repo.load_local_feedstocks(subdir, python_ref, others)
    else:
        aggregate_repo.load_local_feedstocks(subdir, python_ref, others)

    # write package to feedstock mapping file - for reference
    with open(f"./{subdir}/map_to_p.yaml", "w") as f:
        yaml.dump(aggregate_repo.package_to_feedstock_path(), f)

    # get feedstock build order
    blts_buildout = aggregate_repo.get_build_order([], packages)

    # write stage build order
    stages = [
        list(result) for key, result in groupby(blts_buildout, key=lambda f: f.weight)
    ]
    build_order = { "python" : python_ref, "stages" : {} }
    for i, stage in enumerate(stages):
        stage_id = f"{i:03}"
        build_order["stages"][stage_id] = []
        for feedstock in stage:
            build_order["stages"][stage_id].append(feedstock.path)
    with open(f"./{subdir}/blts_{subdir}_build_order.yaml", "w") as f:
        yaml.dump(build_order, f)

if __name__ == "__main__":

    gen_blts_build_order(
        "/Users/cbousseau/work/recipes/aggregate/",
        "linux-64",
        "3.10",
        [
            "python",
            "numpy",
            "scikit-learn",
            "pandas",
            "scikit-image",
            "xgboost",
            "scipy",
            "conda",
            "conda-build",
        ]
    )

    gen_blts_build_order(
        "/Users/cbousseau/work/recipes/aggregate/",
        "osx-arm64",
        "3.10",
        [
            "python",
            "numpy",
            "scikit-learn",
            "pandas",
            "scikit-image",
            "xgboost",
            "scipy",
            "conda",
            "conda-build",
        ]
    )


import click
from percy.commands.recipe import recipe
from percy.commands.aggregate import aggregate


# root click group
@click.group()
def cli():
    pass


# add subcommands
cli.add_command(recipe)
cli.add_command(aggregate)


@cli.command()
def tabcompletion():
    """Print out shell commands for setting up shell completion.

    For local shell use:
        eval "`percy tabcompletion`"

    Add to shell init (i.e. .bashrc, .zshrc) for persistent setup.
    """
    import os

    shell = os.environ["SHELL"]

    if not shell:
        raise click.ClickException("Must set SHELL environment variable")

    if shell.endswith("zsh"):
        os.system("_PERCY_COMPLETE=zsh_source percy")
    elif shell.endswith("bash"):
        os.system("_PERCY_COMPLETE=bash_source percy")
    else:
        raise click.ClickException(f"Unknown SHELL {shell}")


if __name__ == "__main__":
    cli()



""" Copy 311 packages to /www/pkgs/main
"""

import argparse
from pathlib import Path
import requests
import json
import logging

import grp
import os
import pwd
import shutil
import sys
from tempfile import NamedTemporaryFile
from typing import DefaultDict, List

ZEUS_DEST = Path("/www/pkgs/main")
DEFAULT_PERMISSIONS = 0o664


# A custom formatter for our logging purposes.
class CustomFormatter(logging.Formatter):
    """Custom logging format that adds color to different log levels."""

    blue = "\033[0;34m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    boldRed = "\x1b[31;1m"
    reset = "\x1b[0m"
    format_str = "%(name)s - %(levelname)s - %(message)s"

    formats = {
        logging.DEBUG: blue + format_str + reset,
        logging.INFO: blue + format_str + reset,
        logging.WARNING: yellow + format_str + reset,
        logging.ERROR: red + format_str + reset,
        logging.CRITICAL: boldRed + format_str + reset,
    }

    def format(self, record):
        formatter = logging.Formatter(self.formats.get(record.levelno))
        return formatter.format(record)


# Basic logging with our custom formatter.
logger = logging.getLogger("311")
ch = logging.StreamHandler()
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)
logger.setLevel(logging.INFO)


def filter_repodata(subdir, local_dir):
    has_errors = False
    has_warnings = False

    results = {
        "to_copy": set(),
        "present": set(),
        "missing": set(),
        "skip": set(),
    }

    # load defaults repodata
    session = requests.Session()
    repodata_subdir_defaults = None
    url = f"https://repo.anaconda.com/pkgs/main/{subdir}/repodata.json"
    response = session.get(url)
    if response.status_code != 200:
        raise Exception()
    else:
        repodata_subdir_defaults = json.loads(response.text)

    # load local repodata
    local_repodata = local_dir / subdir / "repodata.json"
    with open(local_repodata) as f:
        repodata_subdir_local = json.load(f)
        logger.info("verify local repodata.json against defaults")
        for k, v in repodata_subdir_local["packages"].items():
            if "python >=3.11,<3.12.0a0" in v["depends"]:
                if k in repodata_subdir_defaults["packages"] or not any(
                    [
                        m["name"] == v["name"]
                        and m["version"] == v["version"]
                        and m["build_number"] == v["build_number"]
                        and ("python >=3.11,<3.12.0a0" in m["depends"] or "py311" in p)
                        for p, m in repodata_subdir_defaults["packages"].items()
                    ]
                ):

                    def depends_on_mkl(package):
                        if "blas 1.0 mkl" in package["depends"]:
                            return True
                        return "mkl >=2021.4.0,<2022.0a0" in package["depends"]

                    if subdir == "osx-64" and depends_on_mkl(v):
                        logger.warning(f"Skipping osx-64 mkl related package { k }")
                        has_warnings = True
                        results["skip"].add(k)
                    else:
                        results["to_copy"].add(k)
                else:
                    results["present"].add(k)
            else:
                if k in repodata_subdir_defaults["packages"] or any(
                    [
                        m["name"] == v["name"]
                        and m["version"] == v["version"]
                        and m["build_number"] == v["build_number"]
                        for m in repodata_subdir_defaults["packages"].values()
                    ]
                ):
                    results["present"].add(k)
                else:
                    results["missing"].add(k)

    logger.info("Looking for missing packages")
    if not len(results["missing"]) == 0:
        logger.error(f"Missing packages { results['missing'] }")
        has_errors = True

    # verify every .tar.bz2 has a .conda match
    logger.info("verify cph was called")
    for k in results["to_copy"]:
        fpathtarbz2 = local_dir / subdir / k
        fpathconda = local_dir / subdir / k.replace(".tar.bz2", ".conda")
        if fpathtarbz2.is_file() and not fpathconda.is_file():
            logger.error(f"Package not transmuted to .conda {fpathconda}")
            has_errors = True

    # verify to_copy packages are not in local target folder
    logger.info("verify to_copy packages are not in local target folder")
    filter_to_copy = set()
    filter_copied = set()
    for k in results["to_copy"]:
        fpathtarbz2 = ZEUS_DEST / subdir / k
        fpathconda = ZEUS_DEST / subdir / k.replace(".tar.bz2", ".conda")
        if fpathtarbz2.is_file() or fpathconda.is_file():
            logger.warning(f"Package in target folder! {fpathtarbz2} {fpathconda}")
            has_warnings = True
            filter_copied.add(fpathtarbz2)
            filter_copied.add(fpathconda)
        else:
            fpathtarbz2 = local_dir / subdir / k
            fpathconda = local_dir / subdir / k.replace(".tar.bz2", ".conda")
            filter_to_copy.add(fpathtarbz2)
            filter_to_copy.add(fpathconda)
    results["to_copy"] = filter_to_copy

    return results, has_errors, has_warnings


def copy_failed(msg: str):
    logger.error(msg)
    logger.critical("Sorry captain; try again!")
    sys.exit()


# Actually copy the files.
def copy_files(files: DefaultDict[str, List[str]], dry_run: bool = True):
    if not files:  # if dict is empty it will be false
        logger.critical(
            (
                "Terminating; no files found to transfer. "
                "Better check your inventory captain!"
            )
        )
        sys.exit()

    if dry_run:
        logger.info("Setting transporter for a dry run, captain!")
    if input("Do you wish to continue? (y/n)").lower() != "y":
        logger.info("Terminating - not ready to be beamed up yet, captain?")
        sys.exit()

    user_id = os.getuid()
    user_name = pwd.getpwuid(user_id).pw_name
    group_name = "builders"
    try:
        group_id = grp.getgrnam(group_name).gr_gid
    except Exception:
        copy_failed(f"Failed to obtain id for group '{group_name}'; terminating!")

    for arch, list_of_files in files.items():
        final_path = ZEUS_DEST / arch
        for file in list_of_files:
            try:
                dest_filepath = final_path / os.path.split(file)[1]
                if not dry_run:
                    # doesn't delete but closes the open temporary file on __exit__
                    with NamedTemporaryFile(
                        delete=False,
                        prefix=f"{dest_filepath.name}.",
                        suffix=".scotty",
                        dir=final_path,
                    ) as tf:
                        try:
                            # avoid partially copied packages
                            shutil.copy2(file, tf.name)
                            os.rename(tf.name, dest_filepath)
                        except Exception:
                            os.unlink(tf.name)
                            raise

                logger.info("Copied file %s to %s", file, final_path)
            except Exception:
                copy_failed(f"Could not copy file {file}; terminating!")

            try:
                if not dry_run:
                    os.chmod(dest_filepath, DEFAULT_PERMISSIONS)
                logger.info(
                    "Permissions (%o) set for file %s",
                    DEFAULT_PERMISSIONS,
                    dest_filepath,
                )
            except Exception:
                copy_failed(
                    (
                        f"Could not set permissions ({DEFAULT_PERMISSIONS:o}) "
                        "for file {dest_filepath}; terminating!"
                    )
                )

            try:
                if not dry_run:
                    os.chown(dest_filepath, user_id, group_id)
                logger.info(
                    "Ownership (%s:%s) set for file %s\n",
                    user_name,
                    group_name,
                    dest_filepath,
                )
            except Exception:
                copy_failed(
                    (
                        f"Could not set ownership ({user_name}:{group_name}) "
                        "for file {dest_filepath}; terminating!"
                    )
                )

    if dry_run:
        finish_msg = "Transporter dry run complete"
    else:
        finish_msg = "Everything's up here"

    logger.info("Success - %s! Until next time, captain!", finish_msg)


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="copy what",
        description="Get a list of files to copy",
    )
    parser.add_argument(
        "-s",
        "--subdir",
        type=str,
        help="subdir",
    )

    parser.add_argument(
        "-l",
        "--local_dir",
        type=Path,
        help="local_dir",
    )

    parser.add_argument(
        "-n",
        "--dry-run",
        help="run in dry-run mode, where everything except the actual transfer occurs",
        action="store_true",
    )

    return parser


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()

    results, has_errors, has_warnings = filter_repodata(args.subdir, args.local_dir)
    if has_errors:
        logger.error("Errors were found. Abort...")
        if input("Do you wish to continue? (y/n)").lower() != "y":
            logger.info("Terminating - not ready to be beamed up yet, captain?")
            sys.exit()

    if has_warnings:
        logger.warning("Warning were found.")
        if input("Do you wish to continue? (y/n)").lower() != "y":
            logger.info("Terminating - not ready to be beamed up yet, captain?")
            sys.exit()

    logger.info(f"TO_COPY {results['to_copy']}")
    final_files = {args.subdir: results["to_copy"]}
    copy_files(final_files, args.dry_run)



"""Script to find local packages at lower version than on defaults.

The script considers packages from defaults that depend on python and are
architecture-specific.

It also takes into account exclude packages.

output is outdated_{subdir}.json 
"""

import argparse
import json

import requests
import yaml
from conda.models.version import VersionOrder as vo


def match_lower(mrv, rv, v):
    irv = int(rv) - 1
    hlp = ">={}.{}".format(mrv, irv)
    if v.find(hlp) >= 0:
        return True
    irv = irv - 1
    while irv > -1:
        hlp = ">={}.{}".format(mrv, irv)
        if v.find(hlp) >= 0:
            hlp += ","
            return v.find(hlp) < 0
        hlp = ">{}.{}".format(mrv, irv)
        if v.find(hlp) >= 0:
            hlp += ","
            return v.find(hlp) < 0
        irv = irv - 1
    return False


def depends_on_python(repo_package, python_ref):
    buildstr = repo_package["build"]
    pyver = f"py{python_ref.replace('.','')}"
    if "py_" in buildstr or "pyh" in buildstr or pyver in buildstr:
        return True
    repo_package["depends"]

    hlp = ">=" + python_ref + ",<" + python_ref[0] + "."
    hlp2 = "=" + python_ref
    next_minor = str(int(python_ref.split(".")[-1]) + 1)
    for v in repo_package["depends"]:
        if v.find("python ") >= 0:
            if v.find(hlp) >= 0 or match_lower(python_ref[0], next_minor, v):
                return True
        elif v.find(hlp2) >= 0:
            return True
        elif v == "python":
            return True

    return False


def _is_newer_than(package, version, build_number):
    """Returns True when the package is newer than a version and build_number."""
    if version > package["vo"]:
        return False
    if version < package["vo"]:
        return True
    return int(build_number) < int(package["build_number"])


def compare_repodata(subdir, python_ref, python_target, block_list):
    session = requests.Session()

    # Yields only packages that depend on a python version within a subdir.
    def get_python_packages(arch, python):
        url = f"https://repo.anaconda.com/pkgs/main/{arch}/repodata.json"
        repodata = session.get(url)
        repodata.raise_for_status()
        for pkg in repodata.json()["packages"].values():
            name = pkg["name"]
            if name not in block_list and depends_on_python(pkg, python):
                version = pkg["version"]
                yield name, pkg["build_number"], version, vo(version)

    # Returns the latest versions of each python package.
    def get_latest(arch, python):
        latest: dict = {}
        for name, build, v, version in get_python_packages(arch, python):
            pkg = latest.get(name)
            if not pkg or _is_newer_than(pkg, version, build):
                latest[name] = {
                    "version": v,
                    "vo": version,
                    "build_number": build,
                    "noarch": arch == "noarch",
                }
        return latest

    # Get the published python_ref (defaults) packages.
    pkgs_defaults = get_latest(subdir, python_ref)

    # Remove anything from defaults that is newer in noarch.
    for name, noarch in get_latest("noarch", python_target).items():
        defaults = pkgs_defaults.get(name)
        if not defaults:
            continue
        if _is_newer_than(noarch, defaults["vo"], defaults["build_number"]):
            print(f"Found {name} {noarch}")
            print(f"Removing {name} {defaults}")
            del pkgs_defaults[name]

    # compare local with defaults
    pkgs_local = get_latest(subdir, python_target)
    results = {}
    for local_name, local in pkgs_local.items():
        defaults = pkgs_defaults.get(local_name)
        if defaults and _is_newer_than(defaults, local["vo"], local["build_number"]):
            results[local_name] = {
                "312_version": local["version"],
                "312_build_number": local["build_number"],
                "311_version": defaults["version"],
                "311_build_number": defaults["build_number"],
            }

    # find missing from local
    for defaults_name in pkgs_defaults.keys() - pkgs_local.keys():
        defaults_version = pkgs_defaults[defaults_name]["version"]
        defaults_build_number = pkgs_defaults[defaults_name]["build_number"]
        results[defaults_name] = {
            "312_version": None,
            "312_build_number": None,
            "311_version": defaults_version,
            "311_build_number": defaults_build_number,
        }

    session.close()

    return results


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="gap finder",
        description="Compare two repodata json files",
    )
    parser.add_argument(
        "-s",
        "--subdir",
        type=str,
        help="subdir",
        default="osx-arm64",
    )

    return parser


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()

    block_list = []
    filename = f"{args.subdir}/python_3.11_{args.subdir}_package_list_missing.yaml"
    with open(filename) as file:
        block_list = yaml.safe_load(file)

    results = compare_repodata(args.subdir, "3.11", "3.12", block_list)
    with open(f"./oudated_{args.subdir}.json", "w") as f:
        json.dump(results, f, indent=4)



import sys
from typing import List, TextIO
import yaml
from ruamel.yaml import YAML

from percy.render._renderer import RendererType

# Ruamel configuration
ruamel = YAML(typ="rt")
ruamel.version = (1, 1)
ruamel.indent(mapping=2, sequence=4, offset=2)
ruamel.preserve_quotes = True
ruamel.allow_duplicate_keys = True
ruamel.width = 1000
ruamel.default_flow_style = False


def _dump_render_results_ruamel(render_results: List, out: TextIO = sys.stdout) -> None:
    """Dumps a list of rendered variants of a recipe.

    Args:
        render_results (List[Recipe]): List of rendered variants.
        out (TextIO, optional): Output stream. Defaults to sys.stdout.

    """
    data_to_dump = []
    for render_result in render_results:
        render_dump = render_result.meta
        render_dump["variant"] = render_result.variant_id
        render_dump["renderer"] = str(render_result.renderer)
        for k, v in render_dump["variant"].items():
            render_dump["variant"][k] = list(v)
        data_to_dump.append(render_dump)
    ruamel.dump(data_to_dump, out)


def _dump_render_results_yaml(render_results: List, out: TextIO = sys.stdout) -> None:
    """Dumps a list of rendered variants of a recipe.

    Args:
        render_results (List[Recipe]): List of rendered variants.
        out (TextIO, optional): Output stream. Defaults to sys.stdout.

    """
    FIELDS = [
        "variant",
        "package",
        "source",
        "build",
        "requirements",
        "test",
        "app",
        "outputs",
        "about",
        "extra",
    ]

    class _MetaYaml(dict):
        fields = FIELDS

        def to_omap(self):
            return [(field, self[field]) for field in _MetaYaml.fields if field in self]

    def _represent_omap(dumper, data):
        return dumper.represent_mapping("tag:yaml.org,2002:map", data.to_omap())

    yaml.add_representer(_MetaYaml, _represent_omap)

    class _IndentDumper(yaml.Dumper):
        def increase_indent(self, flow=False, indentless=False):
            return super().increase_indent(flow, False)

        def ignore_aliases(self, data):
            return True

    data_to_dump = []
    for render_result in render_results:
        render_dump = render_result.meta
        render_dump["variant"] = render_result.variant_id
        render_dump["renderer"] = str(render_result.renderer)
        for k, v in render_dump["variant"].items():
            render_dump["variant"][k] = list(v)
        data_to_dump.append(_MetaYaml(render_dump))
    yaml.dump(
        data_to_dump,
        out,
        Dumper=_IndentDumper,
        default_flow_style=False,
        indent=2,
    )


def dump_render_results(render_results: List, out: TextIO = sys.stdout) -> None:
    if render_results and render_results[0].renderer == RendererType.RUAMEL:
        _dump_render_results_ruamel(render_results, out)
    else:
        _dump_render_results_yaml(render_results, out)



"""Recipe renderer.

Not as accurate as conda-build render, but faster and architecture independent.
"""

# TODO: refactor long lines and remove the following linter mute
# ruff: noqa: E501

import sys
import re
import itertools
import logging
from copy import deepcopy
from typing import Any, Dict, List, Set, TextIO
from pathlib import Path
from dataclasses import dataclass, field
from urllib.parse import urlparse
import jsonschema

from percy.render.variants import read_conda_build_config, Variant
from percy.render.exceptions import EmptyRecipe, MissingMetaYaml
import percy.render._renderer as renderer
from percy.render._renderer import RendererType
import percy.render._dumper as dumper


class Recipe:
    """Represents a recipe (meta.yaml) in editable form

    Using conda-build to render recipe is slow and a one-way
    process. We need to be able to load **and** save recipes, which is
    handled by the representation in this class.

    Recipes undergo two manipulation rounds before parsed as YAML:
     1. Selecting lines using ``# [expression]``
     2. Rendering as Jinja2 template

    Args:
        recipe_file (Path): Path to meta.yaml
        variant_id (str): configuration id
        variant (Variant): Variant configuration
        backend (RendererType, optional): Renderer backend.

    Attributes:
        recipe_file (Path): The recipe file
        recipe_dir (Path): The recipe directory
        variant_id (str): configuration id
        selector_dict (Variant): Variant configuration
        meta (Dict[str, Any]): Rendered recipe in as a dictionary
        skip (bool): Whether this variant is skipped.
        packages (Dict[str:Package]): Rendered Packages.
        meta_yaml (List[str]): Lines of the raw recipe file
        orig (Recipe): Original recipe before modifications
    """

    # patch schema
    schema = {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "op": {"enum": ["add", "replace", "remove"]},
                "path": {"type": "string"},
                "match": {"type": "string"},
                "value": {"type": ["string", "array"], "items": {"type": "string"}},
                "description": {"type": "string"},
            },
            "required": [
                "op",
                "path",
            ],
            "additionalProperties": False,
        },
    }

    def __init__(
        self,
        recipe_file: Path,
        variant_id: str = None,
        variant: Variant = None,
        renderer: RendererType = None,
    ):
        """Constructor

        Args:
            recipe_file (Path): Path to meta.yaml
            variant_id (str): configuration id
            variant (Variant): Variant configuration
        """
        #: recipe dir
        if recipe_file:
            self.recipe_file = recipe_file
            self.recipe_dir = recipe_file.parent
        else:
            self.recipe_file = ""
            self.recipe_dir = ""

        #: Selectors configuration
        if not variant_id:
            variant_id = "empty_variant"
        if not variant:
            variant = {}
        self.variant_id = variant_id
        self.selector_dict: Dict[str, Any] = variant

        #: render configuration
        self.renderer = renderer
        if not self.renderer:
            self.renderer = RendererType.PYYAML

        # Filled in by render()
        #: Parsed recipe YAML
        self.meta: Dict[str, Any] = {}
        self.skip = False
        self.packages: Dict[str: Package] = dict()

        # These will be filled in by _load_from_string()
        #: Lines of the raw recipe file
        self.meta_yaml: List[str] = []
        #: Original recipe before modifications (updated by _load_from_string)
        self.orig: Recipe = deepcopy(self)

    @property
    def path(self):
        """Full path to ``meta.yaml``"""
        return self.recipe_file

    @property
    def dir(self):
        """Path to recipe folder"""
        return self.recipe_dir

    def __str__(self) -> str:
        return str(self.recipe_dir)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} "{self.recipe_dir}"'

    def _load_from_string(self, data: str) -> "Recipe":
        """Load and `render` recipe contents from disk

        Args:
            data (str): raw meta.yaml as string.

        Raises:
            EmptyRecipe: The recipe is empty

        Returns:
            Recipe: This recipe object.
        """
        self.meta_yaml = data.splitlines()
        if not self.meta_yaml:
            raise EmptyRecipe(self)
        self.render()
        return self

    @classmethod
    def from_string(
        cls,
        recipe_text: str,
        variant_id: str = None,
        variant: Variant = None,
        return_exceptions: bool = False,
        renderer: RendererType = None,
    ) -> "Recipe":
        """Create new `Recipe` object from string

        Args:
            recipe_text (str): A raw recipe as a string.
            variant_id (str): Variant id
            variant (Variant): Variant configuration.
            return_exceptions (bool, optional): Whether to return exceptions. Defaults to False.
            renderer (RendererType, optional): Renderer backend. Defaults to PYYAML.

        Raises:
            MissingMetaYaml: Missing meta.yaml
            Exception: An exception

        Returns:
            Recipe: A Recipe object.
        """
        try:
            recipe = cls("", variant_id, variant, renderer)
            recipe._load_from_string(recipe_text)
        except Exception as exc:
            if return_exceptions:
                return exc
            raise exc
        recipe._set_original()
        return recipe

    @classmethod
    def from_file(
        cls,
        recipe_fname: str,
        variant_id: str = None,
        variant: Variant = None,
        return_exceptions: bool = False,
        renderer: RendererType = None,
    ) -> "Recipe":
        """Create new `Recipe` object from file

        Args:
            recipe_fname (str): Path to recipe meta.yaml
            variant_id (str): Variant id
            variant (Variant): Variant configuration.
            return_exceptions (bool, optional): Whether to return exceptions. Defaults to False.
            renderer (RendererType, optional): Renderer backend. Defaults to PYYAML.

        Raises:
            MissingMetaYaml: Missing meta.yaml
            Exception: An exception

        Returns:
            Recipe: A Recipe object.
        """
        recipe_fname = Path(recipe_fname)
        recipe = cls(recipe_fname, variant_id, variant, renderer)
        try:
            if recipe_fname.is_file():
                with open(recipe_fname) as text:
                    recipe._load_from_string(text.read())
        except FileNotFoundError:
            exc = MissingMetaYaml(recipe_fname)
            if return_exceptions:
                return exc
            raise exc
        except Exception as exc:
            if return_exceptions:
                return exc
            raise exc
        recipe._set_original()
        return recipe

    def save(self):
        """Save recipe dump to file"""
        with open(self.path, "w", encoding="utf-8") as fdes:
            fdes.write(self.dump())

    def _set_original(self) -> None:
        """Store the current state of the recipe as "original" version"""
        self.orig = deepcopy(self)

    def is_modified(self) -> bool:
        """Has recipe been modified.

        Returns:
            bool: True if recipe has been modified.
        """
        return self.meta_yaml != self.orig.meta_yaml

    def dump(self):
        """Dump recipe content"""
        return "\n".join(self.meta_yaml) + "\n"

    def render(self) -> None:
        """Convert recipe text into data structure

        - create jinja template from recipe content
        - render template
        - parse yaml
        - normalize
        """

        # re-init
        self.meta: Dict[str, Any] = {}
        self.skip = False
        self.packages: Dict[str: Package] = dict()

        # render meta.yaml
        self.meta = renderer.render(
            self.recipe_dir, self.dump(), self.selector_dict, self.renderer
        )

        # should this be skipped?
        bld = self.meta.get("build", {})
        if bld and bld.get("skip", False):
            self.skip = True

        # extract package info
        if not self.skip:
            self._render_packages()

    def _render_packages(self):
        def get_group_from_dev_url(meta, default):
            if not meta:
                return default
            try:
                dev_url = str(meta.get("about", {}).get("dev_url", "") or "").strip()
            except AttributeError:
                dev_url = ""
            if dev_url:
                org = next(iter(urlparse(dev_url).path.lstrip("/").split("/")), "")
                org = str(org or "").strip().lower()
                if org:
                    return org
            else:
                return default

        # read main package deps
        name = self.meta.get("package", {}).get("name", "unknown").strip()
        version = str(self.meta.get("package", {}).get("version", "-1")).strip()
        number = str(dict(self.meta.get("build", {}) or {}).get("number", "0")).strip()
        group = get_group_from_dev_url(self.meta, name)
        path_prefix = ""
        is_noarch = False
        run_exports = []
        ignore_run_exports = []
        main_build = self.meta.get("build", {})
        if main_build:
            is_noarch = main_build.get("noarch", False)
            run_exports = main_build.get("run_exports", [])
            if not run_exports:
                run_exports = []
            run_exports = [
                Dep(i, "build/run_exports")
                for i in run_exports
                if (
                    i is not None
                    and i not in ["noarch", "strong", "weak"]
                    and str(i).strip()
                )
            ]
            ignore_run_exports = main_build.get("ignore_run_exports", [])
            if not ignore_run_exports:
                ignore_run_exports = []
        pkg_reqs = {"build": [], "host": [], "run": [], "run_constrained": []}
        requirements = self.meta.get("requirements", {})
        if requirements is not None:
            if isinstance(requirements, list):
                pkg_reqs["run"].extend(requirements)
            else:
                for s in pkg_reqs.keys():
                    reqs = requirements.get(s, [])
                    if reqs is not None:
                        if isinstance(reqs, list):
                            pkg_reqs[s].extend(reqs)
                        else:
                            pkg_reqs[s].extend([reqs])
            for s in pkg_reqs.keys():
                pkg_reqs[s] = [
                    Dep(i, f"requirements/{s}")
                    for i in pkg_reqs[s]
                    if (i is not None and str(i).strip())
                ]
        test = self.meta.get("test", {})
        test_reqs = []
        if test is not None:
            if isinstance(test, dict):
                reqs = test.get("requires", [])
                if reqs is not None:
                    if isinstance(reqs, list):
                        test_reqs.extend(reqs)
                    else:
                        test_reqs.extend([reqs])
            test_reqs = [
                Dep(i, "test/requires")
                for i in test_reqs
                if (i is not None and str(i).strip())
            ]
        if not self.meta.get("outputs", []):
            # only add if not a multi output recipe
            self.packages[name] = Package(
                self,
                name,
                version,
                number,
                group,
                set(pkg_reqs["build"]),
                set(pkg_reqs["host"]),
                set(pkg_reqs["run"]),
                set(pkg_reqs["run_constrained"]),
                set(run_exports),
                set(ignore_run_exports),
                set(test_reqs),
                is_noarch,
                path_prefix,
            )

        # read output package deps
        outputs = self.meta.get("outputs", [])
        if outputs:
            for n, output in enumerate(outputs):
                name = output.get("name", "")
                version = str(output.get("version", version)).strip()
                group = get_group_from_dev_url(output, group)
                path_prefix = f"outputs/{n}/"
                is_noarch = False
                run_exports = []
                ignore_run_exports = []
                main_build = output.get("build", {})
                if main_build:
                    number = str(main_build.get("number", number) or number).strip()
                    is_noarch = main_build.get("noarch", False)
                    run_exports = main_build.get("run_exports", [])
                    if not run_exports:
                        run_exports = []
                    run_exports = [
                        Dep(i, f"outputs/{n}/run_exports")
                        for i in run_exports
                        if (
                            i is not None
                            and i not in ["noarch", "strong", "weak"]
                            and str(i).strip()
                        )
                    ]
                    ignore_run_exports = main_build.get("ignore_run_exports", [])
                    if not ignore_run_exports:
                        ignore_run_exports = []
                output_pkg_reqs = deepcopy(pkg_reqs)
                requirements = output.get("requirements", {})
                if requirements is not None:
                    if isinstance(requirements, list):
                        output_pkg_reqs["run"].extend(requirements)
                    else:
                        for s in pkg_reqs.keys():
                            reqs = requirements.get(s, [])
                            if reqs is not None:
                                if isinstance(reqs, list):
                                    output_pkg_reqs[s].extend(reqs)
                                else:
                                    output_pkg_reqs[s].extend([reqs])
                    for s in output_pkg_reqs.keys():
                        output_pkg_reqs[s] = [
                            Dep(i, f"outputs/{n}/requirements/{s}")
                            for i in output_pkg_reqs[s]
                            if (i is not None and str(i).strip())
                        ]
                test = output.get("test", {})
                test_reqs = []
                if test is not None:
                    if isinstance(test, dict):
                        reqs = test.get("requires", [])
                        if reqs is not None:
                            if isinstance(reqs, list):
                                test_reqs.extend(reqs)
                            else:
                                test_reqs.extend([reqs])
                    test_reqs = [
                        Dep(i, f"outputs/{n}/test/requires")
                        for i in test_reqs
                        if (i is not None and str(i).strip())
                    ]
                self.packages[name] = Package(
                    self,
                    name,
                    version,
                    number,
                    group,
                    set(output_pkg_reqs["build"]),
                    set(output_pkg_reqs["host"]),
                    set(output_pkg_reqs["run"]),
                    set(output_pkg_reqs["run_constrained"]),
                    set(run_exports),
                    set(ignore_run_exports),
                    set(test_reqs),
                    is_noarch,
                    path_prefix,
                )

    def __getitem__(self, key):
        return self.meta[key]

    def _walk(self, path, noraise=False):
        nodes = [self.meta]
        keys = []
        for key in path.split("/"):
            last = nodes[-1]
            if key.isdigit():
                number = int(key)
                if isinstance(last, list):
                    if noraise and len(last) < number:
                        break
                    nodes.append(last[number])
                    keys.append(number)
                    continue
                if isinstance(last, dict) and number == 0:
                    continue
            if noraise and key not in last:
                break
            nodes.append(last[key])
            keys.append(key)
        return nodes, keys

    def get_raw_range(self, path):
        """Locate the position of a node in the YAML within the raw text

        See also `get_raw()` if you want to get the content of the unparsed
        meta.yaml at a specific key.

        Args:
          path: The "path" to the node. Use numbers for lists ('source/1/url')

        Returns:
          a tuple of first_row, first_column, last_row, last_column
        """
        if not path or not self.renderer == RendererType.RUAMEL:
            if self.meta_yaml:
                return 0, 0, len(self.meta_yaml), len(self.meta_yaml[-1])
            else:
                return 0, 0, 0, 0

        nodes, keys = self._walk(path)
        nodes.pop()  # pop parsed value

        # get the start row/col for the value
        if isinstance(keys[-1], int):
            start_row, start_col = nodes[-1].lc.key(keys[-1])
        else:
            start_row, start_col = nodes[-1].lc.value(keys[-1])

        # getting the end is more complicated, we need to move
        # up the tree to the next item in order until one is not the last
        # item in its collection
        while nodes:
            node = nodes.pop()
            key = keys.pop()
            if isinstance(key, int):
                if key + 1 < len(node):
                    end_row, end_col = node.lc.key(key + 1)
                    break
            else:
                node_keys = list(node.keys())
                if key != node_keys[-1]:
                    next_key = node_keys[node_keys.index(key) + 1]
                    end_row, end_col = node.lc.key(next_key)
                    break
        else:  # reached end of file
            end_row = len(self.meta_yaml) - 1
            end_col = len(self.meta_yaml[end_row])

        # now go backward
        return (start_row, start_col, end_row, end_col)

    def get_raw(self, path):
        """Extracts the unparsed text for a node in the meta.yaml

        This may contain separators and other characters from
        the yaml!

        Args:
          path: Slash-separated path to the node. Numbers can be used
                to access indices in lists. A number '0' is ignored if
                the node is a dict (so 'source/0/url' will work even if
                there is only one url).

        Returns:
          Extracted raw text
        """
        start_row, start_col, end_row, end_col = self.get_raw_range(path)
        if start_row == end_row:
            return self.meta_yaml[start_row][start_col:end_col]

        lines = []
        # first row
        lines.append(self.meta_yaml[start_row][start_col:])
        # middle rows if any
        for row in range(start_row + 1, end_row):
            lines.append(self.meta_yaml[row])
        lines.append(self.meta_yaml[end_row][:end_col])
        return "\n".join(lines).strip()

    def get(self, path: str, default: Any = KeyError) -> Any:
        """Get a value or section from the recipe

        >>> recipe.get('requirements/build')
        ['setuptools]
        >>> recipe.get('source/0/url')
        'https://somewhere'

        The **path** is a ``/`` separated list of dictionary keys to
        be walked in the recipe meta data. Numeric sections in the path
        access list elements. Using ``0`` in the path will get the first
        element in a list or the contents directly if there is no list.
        I.e., `source/0/url` will always get the first url, whether or
        not the source section is a list.

        Args:
          path: Path through YAML
          default: If not KeyError, this value will be returned
                   if the path does not exist in the recipe
        Raises:
          KeyError if no default given and the path does not exist.
        """
        try:
            nodes, keys = self._walk(path)
        except (KeyError, TypeError):
            if default is not KeyError:
                return default
            raise KeyError(f"No '{path}' in Recipe {self}") from None
        res = nodes[-1]
        if default is not KeyError and res is None:
            return default
        return res

    def contains(self, path: str, value: str, default: Any = KeyError) -> bool:
        """Check if a value (string or list) contains a string
        >>> recipe.contains('build/script', 'pip')
        True
        The **path** is a ``/`` separated list of dictionary keys to
        be walked in the recipe meta data. Numeric sections in the path
        access list elements. Using ``0`` in the path will get the first
        element in a list or the contents directly if there is no list.
        I.e., `source/0/url` will always get the first url, whether or
        not the source section is a list.
        Args:
            path: Path through YAML
            value: The string to match
            default: If not KeyError, this value will be returned
                    if the path does not exist in the recipe
        Returns:
            Value match (bool).
        Raises:
            KeyError if no default given and the path does not exist.
        """
        res = self.get(path, default)
        if isinstance(res, str):
            return value in res
        if isinstance(res, list):
            for r in res:
                if value in r:
                    return True
        return False

    def patch(self, operations, increment_build_number=False, evaluate_selectors=True):
        """Patch the recipe given a set of operations.
        Args:
            operations: operations to apply
            increment_build_number: automatically increment the build number of the operations result in changes
            evaluate_selectors: don't evaluate selectors when applying operations
        Returns:
            True if recipe was patched. (bool).
        """
        if self.skip:
            logging.warning(f"Not patching skipped recipe {self.recipe_dir}")
            return False
        if self.renderer != RendererType.RUAMEL:
            self.renderer = RendererType.RUAMEL
            self.render()
        jsonschema.validate(operations, self.schema)
        for op in operations:
            if "@output" not in op["path"]:
                # apply global operation once only
                self._patch(op, evaluate_selectors)
                self.save()
                self.render()
            else:
                # apply package specific operation for all packages
                for package in self.packages.values():
                    opcopy = deepcopy(op)
                    opcopy["path"] = opcopy["path"].replace(
                        "@output/", package.path_prefix
                    )
                    self._patch(opcopy, evaluate_selectors)
                    self.save()
                    self.render()
        if self.is_modified():
            if increment_build_number:
                self._increment_build_number()
                self.save()
                self.render()
            logging.info(f"Patch applied: {self.recipe_dir}")
            return True
        return False

    def _patch(self, operation, evaluate_selectors):
        # read operation parameters
        op = operation["op"]
        path = operation["path"]
        match = operation.get("match", ".*")
        expanded_match = re.compile(
            f"\s+(?P<pattern>{match}[^#]*)(?P<selector>\s*#.*)?"
        )
        value = operation.get("value", [""])
        if value == []:
            value = [""]
        opop = deepcopy(operation)
        opop["path"] = path

        # infer data type
        in_list = False
        add_insert = False
        add_insert_index = 0
        raw_value = self.get(path, "NOPE")
        if op == "remove":
            if raw_value == "NOPE":
                return
        elif op == "replace":
            if raw_value == "NOPE":
                return
            else:
                if isinstance(raw_value, str):
                    if re.search(match, raw_value):
                        parent_path, parent_name = path.rsplit("/", 1)
                        path = parent_path
                        value = [value]
                        expanded_match = re.compile(
                            f"\s+{parent_name}:\s+(?P<pattern>{match}[^#]*)(?P<selector>\s*#.*)?"
                        )
                else:
                    in_list = True
        elif op == "add":
            parent_path, parent_name = path.rsplit("/", 1)
            if raw_value == "NOPE":
                # path not found - add section and return

                # We may be trying to add something that would be removed with
                # the current selector_dict. In that case, better leave it out.
                # Example: adding skip: True # [py<35]
                if evaluate_selectors:
                    if isinstance(value, list):
                        rval = renderer._apply_selector(
                            "\n".join(value), self.selector_dict
                        )
                    else:
                        rval = renderer._apply_selector(value, self.selector_dict)
                    if rval.strip() == "":
                        logging.warning(f"Skipping op due to selector:{opop}")
                        return

                # finding range of direct parent
                # (not doing the leg work of going up the tree if parent is not found)
                try:
                    (start_row, start_col, end_row, _) = self.get_raw_range(parent_path)
                except KeyError:
                    logging.warning(f"Path not found while applying op:{opop}")
                else:
                    # adding value to end of parent
                    # if value is a list, adding as a list to parent
                    # if value is a string, adding as parent: value
                    parent_range = deepcopy(self.meta_yaml[start_row:end_row])
                    parent_insert_index = 0
                    for i, e in reversed(list(enumerate(parent_range))):
                        if e.strip():
                            parent_insert_index = i + 1
                            break
                    if isinstance(value, list):
                        parent_range.insert(
                            parent_insert_index, " " * start_col + f"{parent_name}:"
                        )
                        for val in value:
                            parent_range.insert(
                                parent_insert_index + 1, " " * start_col + f"  - {val}"
                            )
                    else:
                        parent_range.insert(
                            parent_insert_index,
                            " " * start_col + f"{parent_name}: {value}",
                        )
                    self.meta_yaml[start_row:end_row] = parent_range
                    return
            else:
                # path found - store value to add
                try:
                    # Check to see if the path ends with /<n>.
                    # This means that we need to insert the given value at the given index.
                    # See https://www.rfc-editor.org/rfc/rfc6902.html#section-4.1
                    add_insert_index = int(parent_name)
                    add_insert = True

                    # Get the actual raw value of the array.
                    raw_value = self.get(parent_path, "NOPE")

                    path = parent_path
                    parent_path, parent_name = parent_path.split("/", 1)

                    expanded_match = re.compile("NOPE")
                except ValueError:
                    pass

                if isinstance(raw_value, str):
                    if re.search(match, raw_value):
                        path = parent_path
                        expanded_match = re.compile("NOPE")
                        if isinstance(value, list):
                            value = [parent_name + ": " + val for val in value]
                        else:
                            value = [f"{parent_name}: {value}"]
                else:
                    in_list = True

        # get initial section range
        try:
            (start_row, start_col, end_row, _) = self.get_raw_range(path)
        except KeyError:
            logging.warning(f"Path not found while applying op:{opop}")
            return
        range = deepcopy(self.meta_yaml[start_row:end_row])

        # find matching elements
        match_lines = {}
        for i, line in enumerate(range):
            if not line.lstrip().startswith("#"):
                m = expanded_match.search(line)
                if m:
                    match_lines[i] = m
        match_lines = dict(sorted(match_lines.items(), reverse=True))

        # remove elements and find insert position
        new_range = deepcopy(range)
        insert_index = 0
        index_set = False
        for i, m in match_lines.items():
            new_range.pop(i)
            insert_index = i
            index_set = True

        if add_insert:
            index_set = True
            insert_index = add_insert_index

        range = new_range
        if not index_set:
            for i, e in reversed(list(enumerate(range))):
                if e.strip():
                    insert_index = i + 1
                    break

        # add elements
        if op == "add" or (op == "replace" and match_lines):
            to_insert = set()
            for new_val in value:
                for i, m in match_lines.items():
                    to_insert.add(
                        m.string.replace(m.groupdict()["pattern"], new_val).replace(
                            "#", "  #", 1
                        )
                    )
            if not to_insert:
                to_insert = set(value)
            for new_val in to_insert:
                if in_list:
                    range.insert(
                        insert_index, " " * start_col + f"- {new_val.strip(' -')}"
                    )
                else:
                    range.insert(insert_index, " " * start_col + f"{new_val.strip()}")

        # apply change
        self.meta_yaml[start_row:end_row] = range

    def _increment_build_number(self):
        try:
            build_number = int(self.orig.meta["build"]["number"]) + 1
        except (KeyError, TypeError):
            logging.error(f"No build number found for {self.recipe_dir}")
            return
        patterns = (
            ("(?=\s*?)number:\s*([0-9]+)", "number: {}".format(build_number)),
            (
                '(?=\s*?){%\s*set build_number\s*=\s*"?([0-9]+)"?\s*%}',
                "{{% set build_number = {} %}}".format(build_number),
            ),
            (
                '(?=\s*?){%\s*set build\s*=\s*"?([0-9]+)"?\s*%}',
                "{{% set build = {} %}}".format(build_number),
            ),
        )
        text = "\n".join(self.meta_yaml)
        for pattern, replacement in patterns:
            text = re.sub(pattern, replacement, text)
        self.meta_yaml = text.split("\n")


class Dep:
    """A dependency

    Args:
        raw_dep (str): A dependency string. E.g. "numpy <1.24"
        path (str): Path in the recipe. E.g. outputs/1/requirements/run

    Attributes:
        raw_dep (str): A dependency string. E.g. "numpy <1.24"
        pkg (str): The package part. E.g. "numpy"
        variable (str): The constraint part. E.g. "<1.24"
        path (str): Path in the recipe. E.g. outputs/1/requirements/run
    """

    def __init__(self, raw_dep: str, path: str):
        """A dependency

        Args:
            raw_dep (str): A dependency string. E.g. "numpy <1.24"
        """
        self.raw_dep = str(raw_dep)
        splits = re.split(r"[\s<=>]", self.raw_dep, 1)
        self.pkg = splits[0].strip()
        self.constraint = ""
        if len(splits) > 1:
            self.constraint = splits[1]
        self.path = path

    def __str__(self) -> str:
        return str(self.raw_dep)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} "{self.raw_dep}"'


@dataclass
class Package:
    """A rendered package."""

    recipe: Recipe = None
    name: str = None
    version: str = None
    number: str = None
    group: str = None
    build: Set[Dep] = field(default_factory=set)
    host: Set[Dep] = field(default_factory=set)
    run: Set[Dep] = field(default_factory=set)
    run_constrained: Set[Dep] = field(default_factory=set)
    run_exports: Set[str] = field(default_factory=set)
    ignore_run_exports: Set[str] = field(default_factory=set)
    test: Set[Dep] = field(default_factory=set)
    is_noarch: bool = False
    path_prefix: str = ""
    git_info: object = None

    def __getitem__(self, key: str) -> Any:
        """Access this dataclass attributes through square brackets.

        Args:
            key (str): An attribute name of this data class.

        Returns:
            Any: The value of the attribute.
        """
        return getattr(self, key)

    def get(self, key: str, default=None) -> Any:
        """Access this dataclass attributes through a getter.

        Args:
            key (str): An attribute name of this data class.
            default (_type_, optional): Default value. Defaults to None.

        Returns:
            Any: The value of the attribute.
        """
        try:
            return getattr(self, key, default)
        except KeyError:
            return default

    def has_dep(self, section: str, pkg_name: str) -> bool:
        """Returns true if the package is present in the given section.

        Args:
            section (str): A section. E.g. "build", "host", "run", "run_constrained
            pkg_name (str): A package name.

        Returns:
            bool: True if the package is present in the given section.
        """
        return any([pkg_name.lower() == dep.pkg.lower() for dep in self[section]])

    def merge_deps(self, other: "Package"):
        """Merge dependency list of another Package object into this Package.
        This is especially useful to create an aggregate of multiple variants.

        Args:
            other (Package): The other Package.
        """
        self.build.update(other.build)
        self.host.update(other.host)
        self.run.update(other.run)
        self.run_constrained.update(other.run_constrained)
        self.run_exports.update(other.run_exports)
        self.ignore_run_exports.update(other.ignore_run_exports)


def render(
    recipe_path: Path,
    subdir: List[str] = None,
    python: List[str] = None,
    others: Dict[str, Any] = None,
    return_exceptions: bool = False,
    renderer: RendererType = None,
) -> List[Recipe]:
    """Render a recipe

    Args:
        recipe_path (Path): Path to a recipe.
        subdir (List[str], optional): A list of subdir to render for. E.g. ["linux-64", "win-64"]. Defaults to None to render all subdirs.
        python (List[str], optional): A list of python version to render for. E.g. ["3.10", "3.11"]. Defaults to None to render all python.
        others (Dict[str,Any], optional): Additional variants configuration. E.g. {"blas_impl" : "openblas"} Defaults to None.
        return_exceptions (bool, optional): Whether to handle errors as exceptions. Defaults to False.
        renderer (RendererType, optional): Renderer backend. Defautls to PYYAML.

    Returns:
        List[Recipe]: A list of rendered Recipe, one per variant.
    """

    # gather all possible variants
    if others is None:
        others = {"r_implementation": "r-base", "rust_compiler": "rust"}
    if renderer != RendererType.CONDA:
        variants = read_conda_build_config(recipe_path, subdir, python, others)
    else:
        variants = []
        if not python:
            python = ["3.8", "3.9", "3.10", "3.11"]
        for s, p in list(itertools.product(subdir, python)):
            variant = {"subdir": s, "python": [p]}
            variant.update(others)
            variant_id = {"subdir": [s], "python": [p]}
            variants.append((variant_id, variant))

    # render for each variant and combine similar results
    render_results = []
    for variant_id, variant in variants:
        r = Recipe.from_file(
            recipe_path, variant_id, variant, return_exceptions, renderer
        )
        if match := next((x for x in render_results if r.meta == x.meta), None):
            for key, value in r.variant_id.items():
                match.variant_id.get(key, set()).update(value)
        else:
            render_results.append(r)
    return render_results


def dump_render_results(render_results: List[Recipe], out: TextIO = sys.stdout) -> None:
    """Dumps a list of rendered variants of a recipe.

    Args:
        render_results (List[Recipe]): List of rendered variants.
        out (TextIO, optional): Output stream. Defaults to sys.stdout.

    """
    dumper.dump_render_results(render_results, out)



""" Script to find issues in aggregate pinned feedstocks.
"""

import percy.render.aggregate as aggregate
import argparse
from pathlib import Path
import yaml
import requests
import json
import itertools
from conda.models.version import VersionOrder

def get_repodata_package_list(subdir):

    session = requests.Session()

    pkgs = {}

    repodata_subdir = None
    url = f"https://repo.anaconda.com/pkgs/main/{subdir}/repodata.json"
    response = session.get(url)
    if response.status_code != 200:
        raise Exception()
    else:
        repodata_subdir = json.loads(response.text)
    for v in repodata_subdir["packages"].values():
        if v["name"] in pkgs:
            continue
        if VersionOrder(pkgs[v["name"]]["version"]) < VersionOrder(v["version"]):
            pkgs[v["name"]] = { "version": v["version"], "noarch": False }
    
    repodata_noarch = None
    url = "https://repo.anaconda.com/pkgs/main/noarch/repodata.json"
    response = session.get(url)
    if response.status_code != 200:
        raise Exception()
    else:
        repodata_noarch = json.loads(response.text)
    for v in repodata_noarch["packages"].values():
        if v["name"] in pkgs:
            continue
        if VersionOrder(pkgs[v["name"]]["version"]) < VersionOrder(v["version"]):
            pkgs[v["name"]] = { "version": v["version"], "noarch": True }

    return pkgs

def find_issues(aggregate_path, subdir, python_ref, issues, excludes):

    # load defaults
    defaults_pkgs = get_repodata_package_list(subdir)

    # load aggregate
    aggregate_repo = aggregate.Aggregate(aggregate_path)
    others = {"r_implementation": "r-base"}
    if subdir.startswith("win-"):
        others["rust_compiler"] = "rust"
    aggregate_repo.load_local_feedstocks(subdir, python_ref, others)

    for name, rendered_pkg in aggregate_repo.packages.items():

        git_name = rendered_pkg.git_info.name

        def get_issue_type(issue_type):
            return issues[issue_type].setdefault(git_name, [])

        def get_issue(issue_type, name, proto):
            return issues[issue_type].setdefault(git_name, {}).setdefault(name, proto)

        # find local feedstocks with cbc run dep not set in host
        run_deps = set([run_dep.pkg for run_dep in rendered_pkg.run])
        host_deps = set([host_dep.pkg for host_dep in rendered_pkg.host])
        for pkg in set(rendered_pkg.recipe.selector_dict).intersection(run_deps):
            #print(name, pkg, host_deps)
            if pkg not in host_deps and pkg not in excludes:
                rec = get_issue_type("missing_host")
                if pkg not in rec:
                    rec.append(pkg)

        # find local feedstocks ignoring pinning of a host dependency
        pb = set(host_deps).intersection(rendered_pkg.ignore_run_exports)
        for pkg in pb.difference(excludes):
            rec = get_issue_type("bad_ignore_run_exports")
            if pkg not in rec:
                rec.append(pkg)

        # find local feedstocks with outdated version compared to defaults
        if name in defaults_pkgs:
            default_version = defaults_pkgs[name]["version"]
            default_is_noarch = defaults_pkgs[name]["noarch"]
            prototype = {
                "aggregate_version": rendered_pkg.version,
                "defaults_version": default_version,
            }
            if rendered_pkg.version == "-1":
                rendered_pkg.version = "0a"
            if VersionOrder(rendered_pkg.version) < VersionOrder(default_version):
                prototype["subdir"] = [subdir]
                rec = get_issue("outdated_local", name, prototype)
                if subdir not in rec["subdir"]:  # dead code?
                    rec["subdir"].append(subdir)
                if rendered_pkg.has_dep("run", "python") and not default_is_noarch:
                    rec = get_issue("outdated_py_local", name, prototype)
                    if subdir not in rec["subdir"]:
                        rec["subdir"].append(subdir)
            elif VersionOrder(rendered_pkg.version) > VersionOrder(default_version):
                rec = get_issue("outdated_defaults", subdir, [])
                if rendered_pkg.name not in rec:
                    prototype["name"] = rendered_pkg.name
                    rec.append(prototype)
        else:
            rec = get_issue_type("not_in_defaults")
            if rendered_pkg.name not in rec:
                rec.append(rendered_pkg.name)
    
    return issues


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="aggregate_issue_finder",
        description="Find issues in aggregate pinned feedstocks.",
    )
    parser.add_argument(
        "-a",
        "--aggregate",
        type=Path,
        help="aggregate path",
        default="/Users/cbousseau/work/recipes/aggregate/",
    )

    return parser


if __name__ == "__main__":

    parser = create_parser()
    args = parser.parse_args()
    
    subdirs = [
        "osx-arm64",
        "osx-64",
        "win-64",
        "linux-64",
        "linux-aarch64",
        "linux-ppc64le",
        "linux-s390x",
    ]

    python = [
        "3.8",
        "3.9",
        "3.10",
        "3.11",
    ]

    issues = {
        "bad_ignore_run_exports" : {},
        "missing_host" : {},
        "outdated_local" : {},
        "outdated_defaults" : {},
        "outdated_py_local" : {},
        "not_in_defaults" : {},
    }
    issues_no_numpy_python = {
        "bad_ignore_run_exports" : {},
        "missing_host" : {},
        "outdated_local" : {},
        "outdated_defaults" : {},
        "outdated_py_local" : {},
        "not_in_defaults" : {},
    }

    for subdir, pyver in itertools.product(subdirs, python):
        find_issues(args.aggregate, subdir, pyver, issues, [])
        find_issues(
            args.aggregate,
            subdir,
            pyver,
            issues_no_numpy_python,
            ["numpy", "python"],
        )

    with open("./issues.yaml", "w") as f:
        yaml.dump(issues, f)
    with open("./issues_no_numpy_python.yaml", "w") as f:
        yaml.dump(issues_no_numpy_python, f)

    print(f"Outdated: { len(issues['outdated_py_local']) }")


"""A representation of aggregate.

May be used to get a rough build order of packages or gather health information.
"""

# TODO: refactor long lines and remove the linter mute below.
# ruff: noqa: E501

from collections import namedtuple
from dataclasses import dataclass
import subprocess
from pathlib import Path
import logging
import configparser
from typing import Any
from multiprocessing import Pool

from percy.render.recipe import render, Package


class PackageNode:
    """A node representing a package"""

    aggregate = None
    nodes = {}
    bootstrap_nodes = {}

    def __init__(
        self,
        parent: "PackageNode",
        package_name: str,
        package: Package,
        is_run: bool,
    ):
        """PackageNode constructor

        Args:
            parent (PackageNode): The parent node.
            package_name (str): The package name.
            package (Package): The corresponding rendered package.
            is_run (bool): If this is coming from a run dependency.
        """
        self.parents = set([parent])
        self.package_name = package_name
        self.package = package
        self.feedstock = self.aggregate.packages[package_name].git_info
        self.is_run = False
        if is_run:
            self.is_run = True
        if parent:
            self.weight = parent.weight + 1
        else:
            self.weight = 0
        self.children = set()
        logging.debug(f"New node {package_name}")

    @classmethod
    def init(cls, aggregate: "Aggregate"):
        """Initialize a dependency tree.

        Args:
            aggregate (Aggregate): The aggregate object.
        """
        PackageNode.aggregate = aggregate
        PackageNode.nodes = {}
        PackageNode.bootstrap_nodes = {}

    @classmethod
    def make_node(
        cls,
        package_name: str,
        walk_up: bool,
        origin_section: str = "run",
        parent: "PackageNode" = None,
    ) -> "PackageNode":
        """Make a node representing a package.

        Args:
            package_name (str): The name of the package.
            walk_up (bool): Whether to create to call make_node on dependencies of the package.
            origin_section (str, optional): If the package was used in a run section or other. Defaults to "run".
            parent (PackageNode, optional): Parent node. Defaults to None.

        Returns:
            PackageNode: The package node.
        """

        node = None
        package_name = package_name.strip()
        is_run = origin_section == "run"

        def check_cycle(package_name, parent, ancestors=[]):
            if parent:
                if package_name == parent.package_name:
                    return [parent.package_name] + ancestors
                else:
                    for p in parent.parents:
                        r = check_cycle(
                            package_name, p, [parent.package_name] + ancestors
                        )
                        if package_name in r:
                            return r
            return ancestors

        package = PackageNode.aggregate.packages.get(package_name, None)
        if not package:
            logging.warning(f"No information for {package_name}")
        else:
            # check cycle
            cycle = check_cycle(package_name, parent)
            if cycle:
                logging.warning(f"CYCLE: {'->'.join(cycle+[package_name])}")

            if package_name in PackageNode.nodes:
                # existing package - update edges
                node = PackageNode.nodes[package_name]
                if not cycle:
                    node.parents.add(parent)
                if is_run:
                    node.is_run = True
                if parent:
                    node._update_weight(parent.weight)
            else:
                # new package
                node = cls(parent, package_name, package, is_run)
                PackageNode.nodes[package_name] = node
                if walk_up:
                    node._walk_up_requirements()

            if parent:
                # update parent edge
                if not cycle:
                    parent.children.add(node)
                else:
                    PackageNode.bootstrap_nodes[package_name] = node

        return node

    def _update_weight(self, base_weight):
        if self.weight < (base_weight + 1):
            self.weight = base_weight + 1
            for child in self.children:
                child._update_weight(self.weight)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "{}".format(self.package_name)

    def __hash__(self):
        return hash((self.package_name))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.package_name == other.package_name

    def _walk_up_requirements(self):
        for section in ["build", "host", "run"]:
            for dep in self.package[section]:
                PackageNode.make_node(dep.pkg, True, section, self)


@dataclass
class Feedstock:
    name: str
    git_url: str
    branch: str
    path: str
    packages: dict[str, Package]
    weight: int


def _render(feedstock_repo, recipe_path, subdir, python, others):
    try:
        rendered_recipes = render(recipe_path, subdir, python, others)
    except Exception as exc:
        logging.error(f"Render issue {feedstock_repo.name} : {exc}")
        rendered_recipes = []
    return feedstock_repo, rendered_recipes


FeedstockGitRepo = namedtuple("FeedstockGitRepo", ["name", "git_url", "branch", "path"])


class Aggregate:
    """An object to handle a repository of feedstocks.

    Args:
        aggregate_path (str): Aggregate local path.

    Attributes:
        local_path (Path): Aggregate path.
        git_url (str): Aggregate git url.
        git_branch (str): Aggregate git branch.
        packages (dict[str,Package]): Rendered packages. (Populated after calling load_local_feedstocks.)
        feedstocks (dict[str,Feedstock]): Feedstocks. (Populated after calling load_local_feedstocks.)
    """

    def __init__(self, aggregate_path: str):
        """Constructor

        Args:
            aggregate_path (str): Aggregate local path.
        """

        # get local aggregate info
        self.local_path = Path(aggregate_path)
        x = subprocess.run(
            ["git config --get remote.origin.url"],
            capture_output=True,
            shell=True,
            encoding="utf-8",
            cwd=self.local_path,
        )
        self.git_url = x.stdout.strip().replace(
            "git@github.com:", "https://github.com/"
        )
        x = subprocess.run(
            ["git rev-parse --abbrev-ref HEAD"],
            capture_output=True,
            shell=True,
            encoding="utf-8",
            cwd=self.local_path,
        )
        self.git_branch = x.stdout.strip()

        # read git submodules info
        self.submodules = {}
        aggregate_gitmodules = self.local_path / ".gitmodules"
        config = configparser.ConfigParser()
        config.read(aggregate_gitmodules)
        for section in config.sections():
            path = config[section].get("url").rsplit(".git")[0][1:]
            name = section.split('"')[1]
            git_url = f"{self.git_url.rsplit('/')[0]}/{name}.git"
            git_branch = config[section].get("branch", "main")
            self.submodules[name] = FeedstockGitRepo(name, git_url, git_branch, path)

        # packages, feedstocks and groups
        self.packages: dict[str:Package] = {}
        self.feedstocks: dict[str:Feedstock] = {}
        self.groups: dict[str : list[Feedstock]] = {}

    def _get_feedstock_git_repo(self, feedstock_path_rel: Path) -> Feedstock:
        """Get Feedsotck object from feedstock local path.

        Args:
            feedstock_path_rel (Path): Feedsotck local path.

        Returns:
            Feedstock: Feedstock representation.
        """
        feedstock = self.submodules.get(feedstock_path_rel.name, None)
        if feedstock:
            return feedstock
        else:
            return FeedstockGitRepo(
                feedstock_path_rel.name,
                self.git_url,
                self.git_branch,
                f"./{str(feedstock_path_rel)}",
            )

    def load_local_feedstocks(
        self,
        subdir: str = "linux-64",
        python: str = "3.10",
        others: dict[str, Any] = None,
    ) -> dict[str, Package]:
        """Load aggregate feedstocks.

            Only load feedstocks being on master or main branch.
            If feedstocks are already loaded, rendered packages will be merged.

            This populates attributes packages and feedstocks.

        Args:
            subdir (str, optional): The subdir for which to load the feedstocks. Defaults to "linux-64".
            python (str, optional): The python version for which to load the feedstocks. Defaults to "3.10".
            others (dict[str, Any], optional): A variant dictionary. E.g. {"blas_impl" : "openblas"} Defaults to None.

        Returns:
            dict[str, Package]: Rendered packages contained in aggregate (also available as aggregate packages attribute).
        """

        if others is None:
            others = {"r_implementation": "r-base"}
            if subdir == "win-64":
                others["rust_compiler"] = "rust"
            else:
                others["blas_impl"] = "openblas"

        aggregate_path = self.local_path
        to_render = []
        for recipe_path in aggregate_path.glob("**/meta.yaml"):
            # base feedstock info
            feedstock_path_rel = recipe_path.relative_to(aggregate_path).parent
            if feedstock_path_rel.name == "recipe":
                feedstock_path_rel = feedstock_path_rel.parent
            feedstock_name = feedstock_path_rel.parts[0]
            feedstock_repo = self._get_feedstock_git_repo(feedstock_path_rel)

            # attempt to process only the latest feedstocks
            if (
                feedstock_repo.branch
                not in (
                    "master",
                    "main",
                )
                and not feedstock_repo.git_url.endswith("aggregate.git")
                and not feedstock_name == f"python-{python}-feedstock"
            ):
                logging.warning(
                    f"Skipping feedstock {feedstock_name} pinned to branch {feedstock_repo.branch} ({feedstock_repo.git_url})"
                )
                continue
            if feedstock_name.startswith("ctng-compilers-"):
                logging.warning(
                    f"Skipping feedstock {feedstock_name} now replaced by binutils-feedstock and gcc_toolchain-toolchain"
                )
                continue
            if (
                "_cos6_" in feedstock_name
                or "_cos7_" in feedstock_name
                or "_amzn2_" in feedstock_name
            ):
                logging.warning(f"Skipping cdt {feedstock_name}")
                continue

            # add to render list
            to_render.append((feedstock_repo, recipe_path, subdir, python, others))

        # render recipes
        with Pool() as pool:
            results = pool.starmap(_render, to_render)
            for feedstock_repo, rendered_recipes in results:
                for recipe in rendered_recipes:
                    # read recipe packages info
                    if recipe and not recipe.skip:
                        for name, rendered_pkg in recipe.packages.items():
                            rendered_pkg.git_info = feedstock_repo
                            if name in self.packages:
                                logging.debug(
                                    f"Merging deps of other variant {recipe.variant_id} : {feedstock_repo.name} : {name}"
                                )
                                self.packages[name].merge_deps(rendered_pkg)
                            else:
                                self.packages[name] = rendered_pkg

        # process run_exports
        for rendered_pkg in self.packages.values():
            for dep in rendered_pkg.host:
                if dep.pkg in self.packages:
                    for run_export in self.packages[dep.pkg].run_exports:
                        if (
                            "*" not in rendered_pkg.ignore_run_exports
                            and run_export not in rendered_pkg.ignore_run_exports
                        ):
                            rendered_pkg.run.add(run_export)

        # fill feedstocks
        for name, rendered_pkg in self.packages.items():
            v = self.feedstocks.setdefault(
                rendered_pkg.git_info.name,
                Feedstock(
                    rendered_pkg.git_info.name,
                    rendered_pkg.git_info.git_url,
                    rendered_pkg.git_info.branch,
                    rendered_pkg.git_info.path,
                    dict(),
                    0,
                ),
            )
            v.packages[name] = rendered_pkg

        # fill groups
        for name, feedstock in self.feedstocks.items():
            v = self.groups.setdefault(
                next(iter(feedstock.packages.values())).group, []
            )
            v.append(feedstock)

        return self.packages

    def package_to_feedstock_path(self) -> dict[str, str]:
        """Returns a mapping of package name to feedstock path.

        Returns:
            dict[str, str]: mapping of package name to feedstock path.
        """
        package_to_feedstock = {}
        for name, package in self.packages.items():
            package_to_feedstock.setdefault(name, []).append(package.git_info.path)
        return dict(sorted(package_to_feedstock.items()))

    def _build_order(self) -> dict[str, Feedstock]:
        """Computes a feedstock build order.

        Returns:
            dict[str, Feedstock]: An ordered dictionary of Feedstock.
        """
        # feedstock build order
        feedstocks = {}
        for pkg, node in sorted(
            PackageNode.nodes.items(),
            key=lambda x: (1.0 / (x[1].weight + 1), x[1].feedstock),
        ):
            v = feedstocks.setdefault(
                node.feedstock,
                Feedstock(
                    node.feedstock.name,
                    node.feedstock.git_url,
                    node.feedstock.branch,
                    node.feedstock.path,
                    dict(),
                    0,
                ),
            )
            v.weight = max(v.weight, node.weight)
            v.packages[pkg] = self.packages[pkg]
        return sorted(feedstocks.values(), key=lambda x: (1.0 / (x.weight + 1), x.name))

    def get_build_order(
        self,
        target_groups: list[str] = [],
        target_feedstocks: list[str] = [],
        target_packages: list[str] = [],
        drop_noarch: bool = False,
        no_upstream: bool = False,
    ) -> dict[str, Feedstock]:
        """Creates a Feedstock builder order based on a list of leaf packages.

        Args:
            target_groups (list[str], optional): List of target groups.
            target_feedstocks (list[str], optional): List of target feedstocks.
            target_packages (list[str]): List of leaf package names.
            drop_noarch (bool, optional): Whether to drop noarch packages. Defaults to False.
            no_upstream (bool, optional): Whether to drop unspecified packages. Defaults to False.

        Returns:
            dict[str, Feedstock]: An ordered dictionary of Feedstock.
        """
        PackageNode.init(self)

        # expand target_packages with packages from target_feedstocks and target_groups
        target_packages = list(target_packages)
        target_feedstocks = list(target_feedstocks)
        target_groups = list(target_groups)
        for group in target_groups:
            if group in self.groups:
                target_feedstocks.extend([f.name for f in self.groups[group]])
        for feedstock in target_feedstocks:
            if feedstock in self.feedstocks.keys():
                target_packages.extend(self.feedstocks[feedstock].packages.keys())

        # build graph walking up from packages
        for pkg in target_packages:
            PackageNode.make_node(pkg, True)

        # drop graph nodes not having package as a run dependency
        if no_upstream:
            PackageNode.nodes = {
                k: v for k, v in PackageNode.nodes.items() if k in target_packages
            }

        # drop noarch graph nodes
        if drop_noarch:
            PackageNode.nodes = {
                k: v for k, v in PackageNode.nodes.items() if not v.package.is_noarch
            }

        # feedstock build order
        return self._build_order()

    def get_depends_build_order(
        self,
        target_groups: list[str] = [],
        target_feedstocks: list[str] = [],
        target_packages: list[str] = [],
        package_allowlist: list[str] = [],
        feedstock_blocklist: list[str] = [],
        drop_noarch: bool = False,
    ) -> dict[str, Feedstock]:
        """Creates a Feedstock builder order based on packages having target as a dependency.

        Args:
            target_groups (list[str], optional): List of target groups.
            target_feedstocks (list[str], optional): List of target feedstocks.
            target_packages (list[str], optional): List of target packages.
            package_allowlist (list[str], optional): List of package names to consider. Defaults to [].
            feedstock_blocklist (list[str], optional): List of feedstock names to exclude. Defaults to [].
            drop_noarch (bool, optional): Whether to drop noarch packages. Defaults to False.

        Returns:
            dict[str, Feedstock]: An ordered dictionary of Feedstock.
        """

        PackageNode.init(self)

        # expand target_packages with packages from target_feedstocks and target_groups
        target_packages = list(target_packages)
        target_feedstocks = list(target_feedstocks)
        target_groups = list(target_groups)
        for group in target_groups:
            if group in self.groups:
                target_feedstocks.extend([f.name for f in self.groups[group]])
        for feedstock in target_feedstocks:
            if feedstock in self.feedstocks.keys():
                target_packages.extend(self.feedstocks[feedstock].packages.keys())
        logging.info(
            f"get_depends_build_order groups:{target_groups} feedstocks:{target_feedstocks} packages:{target_packages}"
        )

        # build graph with packages having package as a dependency
        for target in target_packages:
            for name, package in self.packages.items():
                if (
                    not package_allowlist or name in package_allowlist
                ) and package.git_info.name not in feedstock_blocklist:
                    for dep in package.run:
                        if dep.pkg == target:
                            PackageNode.make_node(name, True, "run")

        # drop graph nodes in blocklist or not in allowlist
        PackageNode.nodes = {
            k: v
            for k, v in PackageNode.nodes.items()
            if (not package_allowlist or k in package_allowlist)
            and v.feedstock.name not in feedstock_blocklist
        }

        # drop graph nodes not having package as a run dependency
        PackageNode.nodes = {
            k: v
            for k, v in PackageNode.nodes.items()
            if any([v.package.has_dep("run", target) for target in target_packages])
        }

        # drop noarch graph nodes
        if drop_noarch:
            PackageNode.nodes = {
                k: v for k, v in PackageNode.nodes.items() if not v.package.is_noarch
            }

        # feedstock build order
        return self._build_order()



# ruff: noqa: E501

from pathlib import Path
import click
import os
import functools
import json
import percy.render.recipe


def get_recipe(cmd_line=None):
    # command line has highest precedence
    if cmd_line:
        return cmd_line
    # look through current directory
    path = Path(os.getcwd()) / "recipe" / "meta.yaml"
    if path.is_file():
        return path
    else:
        path_old_style = Path(os.getcwd()) / "meta.yaml"
        if path_old_style.is_file():
            return path_old_style
    return path


def base_options(f):
    @click.option(
        "--backend",
        "-b",
        type=click.Choice(["PYYAML", "RUAMEL", "CONDA"], case_sensitive=False),
        default="PYYAML",
        help="Renderer backend. Defaults to pyyaml.",
    )
    @click.option(
        "--subdir",
        "-s",
        type=str,
        multiple=True,
        default=[
            "linux-64",
            "linux-aarch64",
            "linux-ppc64le",
            "linux-s390x",
            "osx-arm64",
            "osx-64",
            "win-64",
        ],
        help="Architecture. E.g. -s linux-64 -s win-64",
    )
    @click.option(
        "--python",
        "-p",
        type=str,
        multiple=True,
        help="Python version. E.g. -p 3.9 -p 3.10",
    )
    @click.option(
        "--others",
        "-k",
        type=(str, str),
        multiple=True,
        default={},
        help="Additional key values (e.g. -k blas_impl openblas)",
    )
    @functools.wraps(f)
    def wrapper_base_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_base_options


@click.group(short_help="Commands for operating on a recipe.")
@click.option("--recipe", "-r", metavar="FILE", help="Recipe meta.yaml to operate on.")
@click.pass_context
def recipe(ctx, recipe):
    """Commands that operate on a recipe."""
    ctx.ensure_object(dict)
    ctx.obj["recipe_path"] = Path(get_recipe(recipe))


@recipe.command(short_help="Render a recipe")
@click.pass_obj
@base_options
def render(obj, subdir, python, others, backend):
    """Render a recipe."""

    # render recipe
    recipe_path = obj["recipe_path"]
    render_results = percy.render.recipe.render(
        recipe_path,
        subdir,
        python,
        dict(others),
        False,
        percy.render.recipe.RendererType[backend],
    )

    # dump recipe
    percy.render.recipe.dump_render_results(render_results)


@recipe.command(short_help="Check if a recipe is outdated compare to main.")
@click.pass_obj
@base_options
def outdated(obj, subdir, python, others, backend):
    """Render a recipe."""

    # render recipe
    recipe_path = obj["recipe_path"]
    render_results = percy.render.recipe.render(
        recipe_path,
        subdir,
        python,
        dict(others),
        False,
        percy.render.recipe.RendererType[backend],
    )

    # load defaults
    print(f"Checking outdated for subdir { subdir[0] }")
    defaults_pkgs = percy.repodata.repodata.get_latest_package_list(subdir[0], True)

    # compare package with defaults
    for recipe in render_results:
        if subdir[0] in recipe.variant_id["subdir"]:
            for name, package in recipe.packages.items():
                result = percy.repodata.repodata.compare_package_with_defaults(
                    package, defaults_pkgs
                )
                if not result:
                    print("OK")
                else:
                    print(f"Outdated: {name} {result}")


@recipe.command(short_help="Patch a recipe")
@click.pass_obj
@base_options
@click.option(
    "--increment_build_number",
    type=bool,
    is_flag=True,
    show_default=True,
    default=False,
    multiple=False,
    help="Increment build number",
)
@click.argument("patch_file", metavar="FILE")
def patch(obj, subdir, python, others, backend, increment_build_number, patch_file):
    """Patch a recipe. Takes a patch file as input, with content like:

    \b
    [
        {"op":"remove","path":"@output/build/noarch"},
        {"op":"add","path":"@output/build/skip","value":["True  # [py<37]"]},
        {"op":"replace","path":"@output/build/script","match":"pip install(?!=.*--no-build-isolation).*","value":["pip install . --no-deps --no-build-isolation --ignore-installed --no-cache-dir -vvv"]},
        {"op":"replace","path":"@output/requirements/host","match":"cython( 0.29.[^\\s]+)?","value":["cython 0.29"]},
        {"op":"replace","path":"@output/requirements/host","match":"numpy( .*)?","value":["numpy {{numpy}}"]},
        {"op":"replace","path":"@output/requirements/run","match":"numpy( .*)?","value":["{{ pin_compatible('numpy') }}"]},
        {"op":"add","path":"@output/requirements/host","match":"python","value":["python"]},
        {"op":"add","path":"@output/requirements/run","match":"python","value":["python"]},
        {"op":"add","path":"@output/requirements/host","match":"setuptools","value":["setuptools"]},
        {"op":"add","path":"@output/requirements/host","match":"wheel","value":["wheel"]},
        {"op":"add","path":"@output/test/requires","match":"pip","value":["pip"]},
        {"op":"add","path":"@output/test/commands","match":"pip check","value":["pip check"]}
    ]
    """

    # render recipe
    recipe_path = obj["recipe_path"]
    render_results = percy.render.recipe.render(
        recipe_path,
        subdir,
        python,
        dict(others),
        False,
        percy.render.recipe.RendererType.RUAMEL,
    )

    # patch recipe
    recipe = next(iter(render_results))
    with open(patch_file) as p:
        recipe.patch(json.load(p), increment_build_number)
    print("Done")



""" Patch a recipe

"""

from pathlib import Path
import logging
import re
import copy
from typing import Any
from ruamel.yaml import YAML

yaml = YAML(typ="jinja2")
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.preserve_quotes = True
yaml.allow_duplicate_keys = True
yaml.width = 1000


logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Recipe:
    def __init__(self, recipe_path):
        self.load(recipe_path)
        self._set_original()

    def load(self, recipe_path):
        self.path = recipe_path
        with open(recipe_path) as text:
            self.meta_yaml = text.read().splitlines()
        with open(recipe_path) as fp:
            self.meta = yaml.load(fp)
            self.packages = {}
            if not self.meta.get("outputs", []):
                name = self.meta.get("package", {}).get("name", "unknown").strip()
                self.packages[name] = ""
            else:
                outputs = self.meta.get("outputs", [])
                for n, output in enumerate(outputs):
                    name = output.get("name", "")
                    self.packages[name] = f"outputs/{n}/"

    def reload(self):
        self.load(self.path)

    def dump(self):
        """Dump recipe content"""
        return "\n".join(self.meta_yaml) + "\n"

    def save(self):
        """Save recipe dump to file"""
        with open(self.path, "w", encoding="utf-8") as fdes:
            fdes.write(self.dump())

    def _set_original(self) -> None:
        """Store the current state of the recipe as "original" version"""
        self.orig = copy.deepcopy(self)

    def is_modified(self) -> bool:
        """Has recipe been modified.

        Returns:
            bool: True if recipe has been modified.
        """
        return self.meta_yaml != self.orig.meta_yaml

    def _walk(self, path, noraise=False):
        nodes = [self.meta]
        keys = []
        for key in path.split("/"):
            last = nodes[-1]
            if key.isdigit():
                number = int(key)
                if isinstance(last, list):
                    if noraise and len(last) < number:
                        break
                    nodes.append(last[number])
                    keys.append(number)
                    continue
                if isinstance(last, dict) and number == 0:
                    continue
            if noraise and key not in last:
                break
            nodes.append(last[key])
            keys.append(key)
        return nodes, keys

    def get_raw_range(self, path):
        """Locate the position of a node in the YAML within the raw text

        See also `get_raw()` if you want to get the content of the unparsed
        meta.yaml at a specific key.

        Args:
          path: The "path" to the node. Use numbers for lists ('source/1/url')

        Returns:
          a tuple of first_row, first_column, last_row, last_column
        """
        if not path:
            if self.meta_yaml:
                return 0, 0, len(self.meta_yaml), len(self.meta_yaml[-1])
            else:
                return 0, 0, 0, 0

        nodes, keys = self._walk(path)
        nodes.pop()  # pop parsed value

        # get the start row/col for the value
        if isinstance(keys[-1], int):
            start_row, start_col = nodes[-1].lc.key(keys[-1])
        else:
            start_row, start_col = nodes[-1].lc.value(keys[-1])

        # getting the end is more complicated, we need to move
        # up the tree to the next item in order until one is not the last
        # item in its collection
        while nodes:
            node = nodes.pop()
            key = keys.pop()
            if isinstance(key, int):
                if key + 1 < len(node):
                    end_row, end_col = node.lc.key(key + 1)
                    break
            else:
                node_keys = list(node.keys())
                if key != node_keys[-1]:
                    next_key = node_keys[node_keys.index(key) + 1]
                    end_row, end_col = node.lc.key(next_key)
                    break
        else:  # reached end of file
            end_row = len(self.meta_yaml) - 1
            end_col = len(self.meta_yaml[end_row])

        # now go backward
        return (start_row, start_col, end_row, end_col)

    def get(self, path: str, default: Any = KeyError) -> Any:
        """Get a value or section from the recipe

        >>> recipe.get('requirements/build')
        ['setuptools]
        >>> recipe.get('source/0/url')
        'https://somewhere'

        The **path** is a ``/`` separated list of dictionary keys to
        be walked in the recipe meta data. Numeric sections in the path
        access list elements. Using ``0`` in the path will get the first
        element in a list or the contents directly if there is no list.
        I.e., `source/0/url` will always get the first url, whether or
        not the source section is a list.

        Args:
          path: Path through YAML
          default: If not KeyError, this value will be returned
                   if the path does not exist in the recipe
        Raises:
          KeyError if no default given and the path does not exist.
        """
        try:
            nodes, keys = self._walk(path)
        except (KeyError, TypeError):
            if default is not KeyError:
                return default
            raise KeyError(f"No '{path}' in Recipe {self}") from None
        res = nodes[-1]
        if default is not KeyError and res is None:
            return default
        return res

    def patch(self, operations, increment_build_number=False):
        for op in operations:
            if "@output" not in op["path"]:
                # apply global operation once only
                self._patch(op)
                self.save()
                self.reload()
            else:
                # apply package specific operation for all packages
                for package_path in self.packages.values():
                    opcopy = copy.deepcopy(op)
                    opcopy["path"] = opcopy["path"].replace("@output/", package_path)
                    self._patch(opcopy)
                    self.save()
                    self.reload()
        if self.is_modified():
            if increment_build_number:
                self._increment_build_number()
                self.save()
                self.render()
            logging.info(f"Patch applied: {self.path}")
            return True
        return False

    def _patch(self, operation):
        # read operation parameters
        op = operation["op"]
        path = operation["path"]
        match = operation.get("match", ".*")
        expanded_match = re.compile(
            f"\s+(?P<pattern>{match}[^#]*)(?P<selector>\s*#.*)?"
        )
        value = operation.get("value", [""])
        if value == []:
            value = [""]
        opop = copy.deepcopy(operation)
        opop["path"] = path

        # infer data type
        in_list = False
        raw_value = self.get(path, "NOPE")
        if op == "remove":
            if raw_value == "NOPE":
                return
        elif op == "replace":
            if raw_value == "NOPE":
                return
            else:
                if not isinstance(raw_value, list):
                    if re.search(match, str(raw_value)):
                        parent_path, parent_name = path.rsplit("/", 1)
                        path = parent_path
                        expanded_match = re.compile(
                            f"\s+{parent_name}:\s+(?P<pattern>{match}[^#]*)(?P<selector>\s*#.*)?"
                        )
                else:
                    in_list = True
        elif op == "add":
            parent_path, parent_name = path.rsplit("/", 1)
            if raw_value == "NOPE":
                # path not found - add section and return

                # finding range of direct parent
                # (not doing the leg work of going up the tree if parent is not found)
                try:
                    (start_row, start_col, end_row, _) = self.get_raw_range(parent_path)
                except KeyError:
                    logging.warning(f"Path not found while applying op:{opop}")
                else:
                    # adding value to end of parent
                    # if value is a list, adding as a list to parent
                    # if value is a string, adding as parent: value
                    parent_range = copy.deepcopy(self.meta_yaml[start_row:end_row])
                    parent_insert_index = 0
                    for i, e in reversed(list(enumerate(parent_range))):
                        if e.strip():
                            parent_insert_index = i + 1
                            break
                    if isinstance(value, list):
                        parent_range.insert(
                            parent_insert_index, " " * start_col + f"{parent_name}:"
                        )
                        for val in value:
                            parent_range.insert(
                                parent_insert_index + 1, " " * start_col + f"  - {val}"
                            )
                    else:
                        parent_range.insert(
                            parent_insert_index,
                            " " * start_col + f"{parent_name}: {value}",
                        )
                    self.meta_yaml[start_row:end_row] = parent_range
                    return
            else:
                # path found - store value to add
                if not isinstance(raw_value, list):
                    if re.search(match, str(raw_value)):
                        path = parent_path
                        expanded_match = re.compile("NOPE")
                        if isinstance(value, list):
                            value = [parent_name + ": " + val for val in value]
                        else:
                            value = [f"{parent_name}: {value}"]
                else:
                    in_list = True

        # get initial section range
        try:
            (start_row, start_col, end_row, _) = self.get_raw_range(path)
        except KeyError:
            logging.warning(f"Path not found while applying op:{opop}")
            return
        range = copy.deepcopy(self.meta_yaml[start_row:end_row])

        # find matching elements
        match_lines = {}
        for i, line in enumerate(range):
            if not line.lstrip().startswith("#"):
                m = expanded_match.search(line)
                if m:
                    match_lines[i] = m
        match_lines = dict(sorted(match_lines.items(), reverse=True))

        # remove elements and find insert position
        new_range = copy.deepcopy(range)
        insert_index = 0
        index_set = False
        for i, m in match_lines.items():
            new_range.pop(i)
            insert_index = i
            index_set = True
        range = new_range
        if not index_set:
            for i, e in reversed(list(enumerate(range))):
                if e.strip():
                    insert_index = i + 1
                    break

        # add elements
        if op == "add" or (op == "replace" and match_lines):
            to_insert = set()
            for new_val in value:
                for i, m in match_lines.items():
                    to_insert.add(
                        m.string.replace(m.groupdict()["pattern"], new_val).replace(
                            "#", "  #", 1
                        )
                    )
            if not to_insert:
                to_insert = set(value)
            for new_val in to_insert:
                if in_list:
                    range.insert(
                        insert_index, " " * start_col + f"- {new_val.strip(' -')}"
                    )
                else:
                    range.insert(insert_index, " " * start_col + f"{new_val.strip()}")

        # apply change
        self.meta_yaml[start_row:end_row] = range

    def _increment_build_number(self):
        build_number = int(self.meta["build"]["number"]) + 1
        patterns = (
            ("(?=\s*?)number:\s*([0-9]+)", "number: {}".format(build_number)),
            (
                '(?=\s*?){%\s*set build_number\s*=\s*"?([0-9]+)"?\s*%}',
                "{{% set build_number = {} %}}".format(build_number),
            ),
            (
                '(?=\s*?){%\s*set build\s*=\s*"?([0-9]+)"?\s*%}',
                "{{% set build = {} %}}".format(build_number),
            ),
        )
        text = "\n".join(self.meta_yaml)
        for pattern, replacement in patterns:
            text = re.sub(pattern, replacement, text)
        self.meta_yaml = text.split("\n")


if __name__ == "__main__":

    operations = [
        {
            "op": "replace",
            "path": "@output/requirements/host",
            "match": "cython\\s+0.29.\\d+",
            "value": ["cython 0.29"]
        },
        {
            "op": "replace",
            "path": "@output/requirements/host",
            "match": "packaging\\s+2\\d.\\d",
            "value": ["packaging 23"]
        },
        {
            "op": "replace",
            "path": "@output/requirements/host",
            "match": "flit-core\\s+3.\\d.\\d",
            "value": ["flit-core 3"]
        },
        {
            "op": "replace",
            "path": "@output/requirements/host",
            "match": "poetry-core\\s+1.5.\\d",
            "value": ["poetry-core 1.5"]
        },
        {
            "op": "replace",
            "path": "@output/requirements/host",
            "match": "hatchling\\s+1.\\d+.\\d",
            "value": ["hatchling 1.18"]
        },
        {
            "op": "replace",
            "path": "@output/requirements/host",
            "match": "setuptools\\s+6\\d.\\d",
            "value": ["setuptools"]
        },
        {
            "op": "replace",
            "path": "@output/requirements/host",
            "match": "setuptools_scm\\s+[\\d\\.]+",
            "value": ["setuptools_scm"]
        }
    ]

    p = Path('.')
    for recipe_path in p.glob("**/recipe/meta.yaml"):
        try:
            recipe = Recipe(str(recipe_path))
            recipe.patch(operations, False)
        except:
            logging.error(f"Failed {recipe_path}")



_DEFAULT_AGGREGATE = "~/work/recipes/aggregate/"
_DEFAULT_PYTHON = ['3.8', '3.9', '3.10', '3.11']
_DEFAULT_SUBDIRS = [
    'linux-64',
    'linux-ppc64le',
    'linux-aarch64',
    'osx-64',
    'osx-arm64',
    'win-64',
    'linux-s390x',
]

def pytest_addoption(parser):
    parser.addoption("--feedstock", action="store", help="feedstock path")
    parser.addoption("--arch", nargs="+", default=_DEFAULT_SUBDIRS)
    parser.addoption("--python", nargs="+", default=_DEFAULT_PYTHON)
    parser.addoption("--channels", nargs="+", default=['default'])
    parser.addoption(
        "--aggregate",
        action="store",
        default=_DEFAULT_AGGREGATE,
        help="directory containing aggregate",
    )
    parser.addoption("--extras", nargs="+", default=[])



import re
import jinja2
import yaml
import contextlib
from typing import List
from enum import Enum

try:
    from ruamel.yaml import YAML
    from ruamel.yaml.parser import ParserError

    has_ruamel = True
except Exception:
    has_ruamel = False
try:
    from conda_build import api
    from conda_build.config import Config

    has_conda_build = True
except Exception:
    has_conda_build = False

from percy.render.exceptions import JinjaRenderFailure, YAMLRenderFailure


class RendererType(Enum):
    PYYAML = 1
    RUAMEL = 2
    CONDA = 3


if has_ruamel:
    # Ruamel configuration
    ruamel = YAML(typ="rt")
    ruamel.version = (1, 1)
    ruamel.allow_duplicate_keys = True
    ruamel.indent(mapping=2, sequence=2, offset=2)
    ruamel.preserve_quotes = True
    ruamel.allow_duplicate_keys = True
    ruamel.width = 1000
    ruamel.default_flow_style = False
    for digit in "0123456789":
        if digit in ruamel.resolver.versioned_resolver:
            del ruamel.resolver.versioned_resolver[digit]

# Pyyaml configuration
try:
    loader = yaml.CLoader
except Exception:
    loader = yaml.Loader


@contextlib.contextmanager
def _stringify_numbers():
    # ensure that numbers are not interpreted as ints or floats.  That trips up versions
    #     with trailing zeros.
    implicit_resolver_backup = loader.yaml_implicit_resolvers.copy()
    for ch in list("0123456789"):
        if ch in loader.yaml_implicit_resolvers:
            del loader.yaml_implicit_resolvers[ch]
    yield
    for ch in list("0123456789"):
        if ch in implicit_resolver_backup:
            loader.yaml_implicit_resolvers[ch] = implicit_resolver_backup[ch]


# Jinja configuration
class _JinjaSilentUndefined(jinja2.Undefined):
    def _fail_with_undefined_error(self, *args, **kwargs):
        class EmptyString(str):
            def __call__(self, *args, **kwargs):
                return ""

        return EmptyString()

    __add__ = (
        __radd__
    ) = (
        __mul__
    ) = (
        __rmul__
    ) = (
        __div__
    ) = (
        __rdiv__
    ) = (
        __truediv__
    ) = (
        __rtruediv__
    ) = (
        __floordiv__
    ) = (
        __rfloordiv__
    ) = (
        __mod__
    ) = (
        __rmod__
    ) = (
        __pos__
    ) = (
        __neg__
    ) = (
        __call__
    ) = (
        __getitem__
    ) = (
        __lt__
    ) = (
        __le__
    ) = (
        __gt__
    ) = (
        __ge__
    ) = (
        __int__
    ) = __float__ = __complex__ = __pow__ = __rpow__ = _fail_with_undefined_error


_jinja_silent_undef = jinja2.Environment(undefined=_JinjaSilentUndefined)


def _apply_selector(data: str, selector_dict: dict) -> List[str]:
    """Apply selectors # [...]

    Args:
        data (str): Raw meta yaml string
        selector_dict (dict): Selector configuration.

    Returns:
        list[str]: meta yaml filtered based on selectors, as a list of string.
    """
    updated_data = []
    for line in data.splitlines():
        if (match := re.search(r"(\s*)[^#].*(#\s*\[([^\]]*)\].*)", line)) is not None:
            cond_str = match.group(3)
            try:
                if not eval(cond_str, None, selector_dict):
                    line = f"{match.group(1)}"
                else:
                    line = line.replace(
                        match.group(2), ""
                    )  # <-- comments sometimes causes trouble in jinja
            except Exception:
                continue
        updated_data.append(line)
    return "\n".join(updated_data)


def _get_template(meta_yaml, selector_dict):
    """Create a Jinja2 template from the current raw recipe"""
    # This function exists because the template cannot be pickled.
    # Storing it means the recipe cannot be pickled, which in turn
    # means we cannot pass it to ProcessExecutors.
    meta_yaml_selectors_applied = _apply_selector(meta_yaml, selector_dict)
    return _jinja_silent_undef.from_string(meta_yaml_selectors_applied)


def render(
    recipe_dir,
    meta_yaml,
    selector_dict,
    renderer_type: RendererType = None,
) -> None:
    """Convert recipe text into data structure

    - create jinja template from recipe content
    - render template
    - parse yaml
    - normalize
    """

    if not renderer_type:
        renderer_type = RendererType.PYYAML

    try:
        #: Variables to pass to Jinja when rendering recipe
        def expand_compiler(lang):
            compiler = selector_dict.get(f"{lang}_compiler", None)
            if not compiler:
                return compiler
            elif renderer_type == RendererType.RUAMEL:
                return f"compiler_{lang}"
            else:
                return f"{compiler}_{selector_dict.get('target_platform', 'win-64')}"

        JINJA_VARS = {
            "unix": selector_dict.get("unix", False),
            "win": selector_dict.get("win", False),
            "PYTHON": selector_dict.get(
                "PYTHON",
                "%PYTHON%" if selector_dict.get("win", False) else "${PYTHON}",
            ),
            "py": int(selector_dict.get("py", "39")),
            "py3k": selector_dict.get("py3k", "0") == "1",
            "py2k": selector_dict.get("py3k", "0") == "0",
            "build_platform": selector_dict.get("target_platform", "win-64"),
            "target_platform": selector_dict.get("target_platform", "win-64"),
            "ctng_target_platform": selector_dict.get("target_platform", "win-64"),
            "cross_target_platform": selector_dict.get("target_platform", "win-64"),
            "ctng_gcc": selector_dict.get("c_compiler_version", "7.3.0"),
            "ctng_binutils": selector_dict.get("c_compiler_version", "2.35"),
            "numpy": selector_dict.get("numpy", "1.16"),
            "np": selector_dict.get("np", "116"),
            "pl": selector_dict.get("pl", "5"),
            "lua": selector_dict.get("lua", "5"),
            "luajit": selector_dict.get("lua", "5")[0] == "2",
            "linux64": selector_dict.get("linux-64", "0") == "1",
            "aarch64": selector_dict.get("aarch64", "0") == "1",
            "ppcle64": selector_dict.get("ppcle64", "0") == "1",
            "cran_mirror": "https://cloud.r-project.org",
            "compiler": expand_compiler,
            "pin_compatible": lambda x, max_pin=None, min_pin=None, lower_bound=None, upper_bound=None: f"{x} x",  # noqa: E501
            "pin_subpackage": lambda x, max_pin=None, min_pin=None, exact=False: f"{x} x",  # noqa: E501
            "cdt": lambda x: f"{x}-cos6-x86_64",
            "os.environ.get": lambda name, default="": "",
            "ccache": lambda name, method="": "ccache",
        }
        render_dict = {**JINJA_VARS, **selector_dict}
        yaml_text = _get_template(meta_yaml, selector_dict).render(render_dict)
    except jinja2.exceptions.TemplateSyntaxError as exc:
        raise JinjaRenderFailure(recipe_dir, message=exc.message, line=exc.lineno)
    except jinja2.exceptions.TemplateError as exc:
        raise JinjaRenderFailure(recipe_dir, message=exc.message)
    except TypeError as exc:
        raise JinjaRenderFailure(recipe_dir, message=str(exc))

    try:
        if renderer_type == RendererType.RUAMEL:
            if not has_ruamel:
                raise YAMLRenderFailure(recipe_dir, message="ruamel unavailable")
            # load yaml with ruamel
            return ruamel.load(yaml_text.replace("\t", " ").replace("%", " "))
        elif renderer_type == RendererType.PYYAML:
            # load yaml with pyyaml
            with _stringify_numbers():
                return yaml.load(
                    yaml_text.replace("\t", " ").replace("%", " "), Loader=loader
                )
        elif renderer_type == RendererType.CONDA:
            if not has_conda_build:
                raise YAMLRenderFailure(recipe_dir, message="conda build unavailable")
            platform, arch = selector_dict.get("subdir").split("-")
            rendered = api.render(
                recipe_dir,
                config=Config(
                    platform=platform,
                    arch=arch,
                ),
                variants=selector_dict,
            )
            return rendered[0][0].meta
        else:
            raise YAMLRenderFailure(recipe_dir, message="Unknown renderer type.")
    except ParserError as exc:
        raise YAMLRenderFailure(recipe_dir, line=exc.problem_mark.line)
